# Licensed under GPL version 3 - see LICENSE.rst
'''Tools for setting up instruments in the Rowland Torus geometry.

This includes an object that represents the Rowland torus itself (imaginatively
called `~marxs.design.rowland.RowlandTorus`), some helper functions used to set
the right parameters for the torus and a few classes that are derived from
`marxs.simulator.ParallelCalculated`, each placing elements such as gratings
or detectors on the Rowland torus. There are many ways to do that, e.g. the
limits can be defined in x,y,z coordiantes or in theta, phi coordinates on the
torus, gratings can be ordered in concentric circles or pack densely in a
rectoangular area etc. At this point, the different classes do not exploit all
these possibilities, they merely give a few of many possible ways to set up an
instrument.

These classes may be generalized in the future.
'''

from __future__ import division

import numpy as np
from scipy import optimize
import transforms3d
from transforms3d.utils import normalized_vector

from ..optics.base import OpticalElement
from ..base import MarxsElement
from ..optics import FlatDetector
from ..math.rotations import ex2vec_fix
from ..math.utils import e2h, h2e, anglediff
from ..simulator import ParallelCalculated
from ..math.geometry import Geometry

__all__ = ['find_radius_of_photon_shell', 'design_tilted_torus',
           'RowlandTorus',
           'ElementPlacementError', 'GratingArrayStructure', 'LinearCCDArray',
           'RectangularGrid', 'RowlandCircleArray']

def find_radius_of_photon_shell(photons, mirror_shell, x, percentile=[1,99]):
    '''Find the radius the photons coming from a single mirror shell have.

    For nested Wolter Type I mirrors the ray of photons reflected from a single
    mirror shell essentially form a cone in space. The tip of the cone is at
    the focal point and the base is at the mirror. There is a certain thickness
    to this cone depending on where exactly on the mirror the individual photon
    was reflected.

    This function takes a photon list of photons after passing through the
    mirror and calculates the radius range that this photon cone covers at a
    specific distance from the focal point. This information can help to design
    the placement of gratings.

    Parameters
    ----------
    photons : `~astropy.table.Table`
        Photon list with position and direction of photons leaving the mirror
    mirror_shell : int
        Select mirror shell to look at (uses column ``mirror_shell`` in
        ``photons`` for filtering).
    x : float
        Distance along the optical axis (assumed to coincide with the x axis
        with focal point at 0).
    percentile : list of floats
        The radius is calculated at the given percentiles. ``50`` would give
        the median radius. The default of ``[1, 99]`` gives a radius range
        excluding extrem outliers such as stray rays scattered into the extreme
        wing of the PSF.

    '''
    p = photons.copy()
    mdet = FlatDetector(position=np.array([x, 0, 0]), zoom=1e8, pixsize=1.)
    p = mdet(p)
    ind = (p['probability'] > 0) & (p['mirror_shell'] == mirror_shell)
    r = np.sqrt(p['det_x'][ind]**2+p['det_y'][ind]**2.)
    return np.percentile(r, percentile)


class RowlandTorus(MarxsElement, Geometry):
    '''Torus with y axis as symmetry axis.

    Note that the origin of the torus is the focal point, which is
    **may or may not** be the same as the center of the torus.

    Parameters
    ----------
    R : float
        Radius of Rowland torus. ``r`` determines the radius of the Rowland
        circle, ``R`` is then used to rotate that circle around the axis of
        symmetry of the torus.
    r : float
        Radius of Rowland circle
    '''

    display = {'color': (1., 0.3, 0.3),
               'opacity': 0.2,
               'shape': 'torus; surface',
               'coo1': np.linspace(0, 2 * np.pi, 60),
               'coo2': np.linspace(0, 2 * np.pi, 60)}

    def __init__(self, R, r, **kwargs):
        self.R = R
        self.r = r
        # super does not work, because the various classes have different signature
        Geometry.__init__(self, kwargs)
        MarxsElement.__init__(self, **kwargs)

    def quartic(self, xyz, transform=True):
        '''Quartic torus equation.

        Roots of this equation are points on the torus.

        Parameters
        ----------
        xyz : np.array of shape (N, 3) or (3)
            Coordinates of points in euklidean space. The quartic is calculated
            for those points.
        transform : bool
            If ``True`` transform ``xyz`` from the global coordinate system
            into the local coordinate system of the torus. If this
            transformation is done in the calling function already, set to
        ``False``.

        Returns
        -------
        q : np.array of shape (N) or scalar
            Quartic at the input location
        '''
        if xyz.shape[-1] != 3:
            raise ValueError('Input coordinates must be defined in Eukledian space.')

        if transform:
            invpos4d = np.linalg.inv(self.pos4d)
            xyz = h2e(np.einsum('...ij,...j', invpos4d, e2h(xyz, 1)))
        return ((xyz**2).sum(axis=-1) + self.R**2. - self.r**2.)**2. - 4. * self.R**2. * (xyz[..., [0,2]]**2).sum(axis=-1)

    def solve_quartic(self, x=None, y=None, z=None, interval=[0, 1], transform=True):
        '''Solve the quartic on the Rowland torus in Cartesian coordinates.

        This method solves the quartic equation for positions on the Rowland
        Torus for cases where two of the Cartesian coordinates are fixed
        (e.g. y and z) and the third one (e.g. x) needs to be computed. This
        function is intended as a convenience for a common use case. In more
        general cases, evaluate the :meth:`RowlandTorus.quartic` and search for
        the roots of that function.

        Parameters
        ----------
        x, y, z : float or None
            Set two of these coordinates to fixed numbers. This method will
            solve for the coordinate set to ``None``.
            x, y, z are defined in the global coordinate system.
        interval : np.array
            [min, max] for the search. The quartic can have up to four
            solutions because a line can intersect a torus in four points and
            this interval must bracket one and only one solution.
        transform : bool
            If ``True`` transform ``xyz`` from the global coordinate system
            into the local coordinate system of the torus. If this
            transformation is done in the calling function already, set to
            ``False``.

        Returns
        -------
        coo : float
            Value of the fitted coordinate.

        '''
        n_Nones = 0
        for i, c in enumerate([x, y, z]):
            if c is None:
                n_Nones +=1
                ind = i
        if n_Nones != 1:
            raise ValueError('Exactly one of the input numbers for x,y,z must be None.')
        # Need to give it a number for vstack to work
        if ind == 0: x = 0.
        if ind == 1: y = 0.
        if ind == 2: z = 0.

        xyz = np.vstack([x,y,z]).T
        def f(val_in):
            xyz[..., ind] = val_in
            return self.quartic(xyz, transform=transform)
        val_out, brent_out = optimize.brentq(f, interval[0], interval[1], full_output=True)
        if not brent_out.converged:
            raise Exception('Intersection with torus not found.')
        return val_out

    def parametric_surface(self, theta, phi, display):
        '''Parametric representation of surface of torus.

        In contrast to `parametric` the input parameters here are 1-d arrays
        and the functions converts that to a rectangular grid.

        Parameters
        ----------
        theta : np.array
            Points on the Rowland circle (specified in rad, where 0 is on the
            optical axis across from the focal point).
        phi : np.array
            ``phi`` rotates the Rowland circle to make up the Rowland torus.

        Returns
        -------
        xyzw : np.array
            Torus coordinates in global homogeneous coordinate system.
        '''
        if (phi.ndim != 1) or (theta.ndim != 1):
            raise ValueError('input parameters have 1-dim shape.')
        theta, phi = np.meshgrid(theta, phi)
        return self.parametric(theta, phi)

    def parametric(self, theta, phi):
        '''Parametric description of points on the torus.

        This is just another way to obtain the shape of the torus, e.g.
        for visualization.

        Parameters
        ----------
        theta : np.array
            Points on the Rowland circle (specified in rad, where 0 is on the
            optical axis across from the focal point).
        phi : np.array
            ``phi`` rotates the Rowland circle to make up the Rowland torus.

        Returns
        -------
        xyzw : np.array
            Torus coordinates in global homogeneous coordinate system.
        '''
        x = (self.R + self.r * np.cos(theta)) * np.cos(phi)
        z = (self.R + self.r * np.cos(theta)) * np.sin(phi)
        y = self.r * np.sin(theta)
        w = np.ones_like(z)
        torus = np.array([x, y, z, w]).T
        return np.einsum('...ij,...j', self.pos4d, torus)

    def xyzw2parametric(self, xyzw, transform=True, intersectvalid=True):
        '''Calculate (theta, phi) coordinates for point on torus surface.

        Parameters
        ----------
        xyzw : np.array of shape (N, 4)
            Coordinates of points on the torus surface in homogeneous
            coordiantes.

        transform : bool
            If ``True`` transform ``xyz`` from the global coordinate system
            into the local coordinate system of the torus. If this
            transformation is done in the calling function already, set to
            ``False``.

        intersectvalid : bool
            When ``r >=R`` the torus can intersect with itself. At these points,
            phi is not unique. If ``intersectvalid`` is true, those points will
            be filled with an arbitrarily chosen valid value (0.), otherwise
            they will be nan.

        Returns
        -------
        theta, phi: np.array
            Parametric representation for all points on the torus.
        '''
        if transform:
            invpos4d = np.linalg.inv(self.pos4d)
            xyzw = np.einsum('...ij,...j', invpos4d, xyzw)

        xyz = h2e(xyzw)

        if not np.allclose(self.quartic(xyz, transform=False) / self.R**4., 0.):
            raise ValueError('Parametric representation is only defined for points on torus surface.')
        # There are two branches and we can use the sign to distinguish between them
        s = np.sign(np.sqrt(xyz[:,0]**2 + xyz[:, 2]**2) - self.R)
        theta = np.arcsin(s * xyz[:, 1] / self.r) + (s < 0) * np.pi

        factor = self.R + self.r * np.cos(theta)
        phi = np.zeros_like(factor)
        ind = factor != 0
        phi[ind] = np.arctan2(xyz[ind, 2] / factor[ind], xyz[ind, 0] / factor[ind])
        if not intersectvalid:
            phi[~factor] = np.nan

        return theta, phi

    def normal_parametric(self, theta, phi):
        '''Return the gradient vector field.

        Following the usual concentions, the vector is pointing outwards
        of the torus volume.

        Parameters
        ----------
        theta, phi : 1-d np.arrays
            Theta and phi coordinates for points on the torus surface.

        Returns
        -------
        gradient : np.array
            Gradient vector field in homogeneous coordinates. One vector
            corresponds to each input point. The shape of ``gradient`` is the
            same as the shape of ``xyz``.
        '''
        if not ((theta.ndim == 1) and (phi.ndim == 1)):
            raise ValueError('theta and phi must be 1-d arrays.')
        normal = np.empty((len(theta), 3))
        normal[:, 0] = np.cos(theta) * np.cos(phi)
        normal[:, 1] = np.sin(theta)
        normal[:, 2] = np.cos(theta) * np.sin(phi)

        return np.einsum('...ij,...j', self.pos4d, e2h(normal, 0))


    def normal(self, xyzw):
        '''Return the gradient vector field.

        Following the usual concentions, the vector is pointing outwards
        of the torus volume.

        Parameters
        ----------
        xyzw : np.array of shape (N, 4)
            Coordinates of points in euklidean space. The quartic is calculated
            for those points. All points need to be on the surface of the torus.

        Returns
        -------
        gradient : np.array
            Gradient vector field in euklidean coordinates. One vector
            corresponds to each input point. The shape of ``gradient`` is the
            same as the shape of ``xyz``.
        '''
        if not len(xyzw.shape) == 2:
            raise ValueError('Shape of input array must be (N, 4).')
        theta, phi = self.xyzw2parametric(xyzw, 1)
        return self.normal_parametric(theta, phi)

    def xyz_from_radiusangle(self, radius, angle, interval):
        '''Get Cartesian coordiantes for radius, angle on the rowland circle.

        y, z are calculated from the radius and angle of polar coordiantes in a
        plane; then x is determined from the condition that the point lies on
        the Rowland circle.  The plane is perpendicular to the optical axis
        that defines the Rowland circle.

        Parameters
        ----------
        radius, angle : float or np.array of shape (n,)
            Polar coordinates in a plane perpendicular to the optical axis
            (where the optical axis is parallel to the x-axis and goes through
            the origin of the `RowlandTorus`.
            ``angle=0`` conicides with the local y-axis.
        interval : np.array
            [min, max] for the search. The quartic can have up to four
            solutions because a line can intersect a torus in four points and
            this interval must bracket one and only one solution.

        Returns
        -------
        xyz : np.array of shape (n, 3)
            Eukledian coordinates in the global coordinate system.
        '''
        y = radius * np.cos(angle)
        z = radius * np.sin(angle)
        x = self.solve_quartic(y=y,z=z, interval=interval, transform=False)
        xyz = np.vstack([x,y,z, np.ones_like(x)]).T
        return h2e(np.einsum('...ij,...j', self.pos4d, xyz))


def design_tilted_torus(f, alpha, beta):
    '''Design a torus with specifications similar to Heilmann et al. 2010

    A `RowlandTorus` is fully specified with the parameters ``r``, ``R`` and
    ``pos4d``.  However, in practice, these numbers might be derived from other
    values.  This function calculates the parameters of a RowlandTorus, based
    on a different set of input values.

    Parameters
    ----------
    f : float
        distance between focal point and on-axis grating. Should be as large as
        possible given the limitations of the spacecraft to increase the
        resolution.
    alpha : float (in radian)
        angle between optical axis and the line (on-axis grating - center of
        Rowland circle).
        A typical value could be twice the blaze angle.
    beta : float (in radian)
        angle between optical axis and the line (on-axis grating - hinge),
        where the hinge is a point on the Rowland circle. The Rowland Torus
        will be constructed by rotating
        the Rowland Circle around the axis (focal point - hinge).
        The center of the Rowland Torus will be the point where the line
        (on-axis grating - center of Rowland circle) intersects the line
        (focal point - hinge).

    Returns
    -------
    R : float
        Radius of Rowland torus. ``r`` determines the radius of the Rowland
        circle, ``R`` is then used to rotate that circle around the axis of
        symmetry of the torus.
    r : float
        Radius of Rowland circle
    pos4d : np.array of shape (4, 4)

    Notes
    -----
    The geometry used here really needs to be explained in a figure.
    However, some notes to explain at least the meaning of the symbols on the
    code are in order:

    - Cat : position of on-axis CAT grating (where the Rowland circle intersects the on-axis beam)
    - H : position of hinge
    - Ct : Center of Rowland Torus
    - F : Focal point on axis (at the origin of the coordinate system)
    - CatH, HF, FCt, etc. : distance between Cat and H, F and Ct, etc.
    - gamma : see sketch.
    '''
    r = f / (2. * np.cos(alpha))
    CatH = r * np.sqrt(2 * (1 + np.cos(2 * (beta - alpha))))
    HF = np.sqrt(f**2 + CatH**2 - 2 * f * CatH * np.cos(beta))
    # If alpha is negative, then everything is "on the other side".
    # The sign of gamma cannot be found through the arccos, so need to set it here
    # with sign(alpha).
    # Another gotcha: np.sign(0) = 0, but we want 1 (or -1)
    gamma = np.arccos(HF / (2 * r)) * (np.sign(alpha) or 1)
    R = f / np.sin(np.pi - alpha - (alpha + gamma)) * np.sin(alpha + gamma) - r
    FCt = f / np.sin(np.pi - alpha - (alpha + gamma)) * np.sin(alpha)
    x_Ct = FCt * np.cos(alpha + gamma)
    z_Ct = 0
    y_Ct = FCt * np.sin(alpha + gamma)
    orientation = transforms3d.axangles.axangle2mat([0, 0, -1], np.pi/2 - alpha - gamma)
    pos4d = transforms3d.affines.compose([x_Ct, y_Ct, z_Ct], orientation, np.ones(3))
    return R, r, pos4d


class ElementPlacementError(Exception):
    pass


class RowlandCircleArray(ParallelCalculated, OpticalElement):
    '''A 1D collection of elements (e.g. CCDs) arranged on a Rowland circle.

    When a `RowlandCircleArray` is initialized, it places a number of elements
    on the Rowland circle. These elements could be any optical element, but the
    most common use for this structure is an array of CCDs capturing a
    spread-out grating spectrum like ACIS-S in Chandra.

    After generation, individual positions can be adjusted by hand by
    editing the attributes `elem_pos` or `elem_uncertainty`.
    See `marxs.simulator.Parallel` for details.

    After any of the `elem_pos`, `elem_uncertainty` or
    `uncertainty` is changed, `generate_elements` needs to be
    called to regenerate the final CCD positions.

    Parameters
    ----------
    rowland : RowlandTorus
    d_element : float
        Size of the edge of each element, which is assumed to be flat and
        square. ``d_element`` can be larger than the actual size of the optical
        element to accommodate a minimum distance between elements from
        mounting structures.
    theta : list of floats
        Angle on the Rowland circle to be covered by detectors.
        For a continuous array of detectors, this is just a list with two
        elements ``[inner, outer]``. However, it is also possible to list more
        than one range in a flat list, to e.g. set one detector in the focus to
        detect the zeroth order and offset others:
        ``[inner_1, outer_1, inner_2, outer_2, ...]``.
    '''

    id_col = 'CCD_ID'

    def __init__(self, rowland, d_element, theta, **kwargs):
        self.rowland = rowland
        if not len(theta) % 2 == 0:
            raise ValueError('radius must be a list of [inner_1, outer_1, inner_2, outer_2, ...].')
        if np.max(np.abs(theta)) > 10:
            raise ValueError('Input angles >> 2 pi. Did you use degrees (radian expected)?')
        self.theta = theta
        self.d_element = d_element
        kwargs['normal_spec'] = self.rowland_normal
        kwargs['parallel_spec'] =  self.rowland.parametric(0., 0.) - self.rowland.parametric(1., 0.)
        kwargs['pos_spec'] = self.xyzwpos

        super(RowlandCircleArray, self).__init__(**kwargs)

    def rowland_normal(self, xyzw):
        return self.rowland.normal(xyzw)

    def xyzwpos(self):
        radii = self.distribute_elements_on_arc()
        return self.rowland.parametric(radii, 0.)

    def max_elements_on_arc(self, theta):
        '''Max number of elements that fit on an arc

        Parameters
        ----------
        theta : list of two floats
            angle range that should be covered by elements

        Returns
        -------
        n : int
            Number of elements needed to cover a given radius segment.
            Elements might reach beyond the limits if the arc length
            is not an integer multiple of the element size.
        '''
        return int(np.ceil(self.rowland.r * (theta[1] - theta[0]) / self.d_element))

    def distribute_elements_on_arc(self):
        '''Distributes elements as evenly as possible along an arc segment.

        Returns
        -------
        theta : np.ndarray
            Theta coordinates of the element *center* positions.
        '''
        theta = []
        for i in range(len(self.theta) // 2):
            bracket = self.theta[2 * i: 2 * i + 2]
            n = self.max_elements_on_arc(bracket)
            theta.append(np.mean(bracket) +
                         np.arange(- n / 2 + 0.5, n / 2 + 0.5) * self.d_element / self.rowland.r)
        return np.hstack(theta)


class LinearCCDArray(ParallelCalculated, OpticalElement):
    '''A 1D collection of elements (e.g. CCDs) arranged on a Rowland circle.

    When a `LinearCCDArray` is initialized, it places a number of elements on
    the Rowland circle. These elements could be any optical element, but the
    most common use for this structure is an array of CCDs capturing a
    spread-out grating spectrum like ACIS-S in Chandra.

    After generation, individual positions can be adjusted by hand by editing
    the attributes `elem_pos` or `elem_uncertainty`. See `Parallel` for
    details.

    After any of the `elem_pos`, `elem_uncertainty` or
    `uncertainty` is changed, `generate_elements` needs to be
    called to regenerate the final CCD positions.

    Parameters
    ----------
    rowland : RowlandTorus
    d_element : float
        Size of the edge of each element, which is assumed to be flat and
        square. (``d_element`` can be larger than the actual size of the
        optical element to accommodate a minimum distance between elements
        from mounting structures.
    x_range: list of 2 floats
        Minimum and maximum of the x coordinate that is searched for an
        intersection  with the torus. A line can intersect a torus in up to
        four points. ``x_range`` specifies the range for the numerical search
        for the intersection point.
    radius : list of floats
        Inner and outer radius as measured in the yz-plane from the center of
        the `LinearCCDArray`. Can be negative to place elements on both sides
        of the center of the `LinearCCDArray`. Elements will be placed
        ``d_element`` apart; if a non-integer number of elements is needed to
        cover the ``radius``, elements will reach beyond the given numbers.
        For a continuous array of detectors, this is just a list with two
        elements ``[r_inner, r_outer]``. However, it is also possible to list
        more than one range in a flat list, to e.g. set one detector in the
        focus to detect the zeroth order and offset others:
        ``[r_inner_1, r_outer_1, r_inner_2, r_outer_2, ...]``.
    phi : floats
        Direction of line through the centers of all elements. :math:`\phi=0`
        is on the positive y axis. Angles are given in radian.
    '''

    id_col = 'CCD_ID'

    def __init__(self, rowland, d_element, x_range, radius, phi, **kwargs):
        self.rowland = rowland
        if not len(radius) % 2 == 0:
            raise ValueError('radius must be a list of [inner_1, outer_1, inner_2, outer_2, ...].')
        radarray = np.array(radius)
        if not np.all(radarray[1::2] > radarray[::2]):
            raise ValueError('Outer radius must be larger than inner radius.')
        self.radius = radius

        if np.max(np.abs(phi)) > 10:
            raise ValueError('Input angles >> 2 pi. Did you use degrees (radian expected)?')
        self.phi = phi
        self.x_range = x_range
        self.d_element = d_element

        if 'normal_spec' not in kwargs.keys():
            kwargs['normal_spec'] = self.rowland_normal

        if 'parallel_spec' not in kwargs.keys():
            radii = self.distribute_elements_on_radius()
            kwargs['parallel_spec'] =  e2h(normalized_vector(self.xyz_from_radiusangle(radii[1], self.phi, self.x_range) - self.xyz_from_radiusangle(radii[0], self.phi, self.x_range)), 0)

        if 'pos_spec' not in kwargs.keys():
            kwargs['pos_spec'] = self.xyzwpos

        super(LinearCCDArray, self).__init__(**kwargs)

    def rowland_normal(self, xyzw):
        return self.rowland.normal(xyzw)

    def xyz_from_radiusangle(self, r, phi, x_range):
        '''Wrap `marxs.design.RowlandTorus.xyz_from_radiusangle` for better error message'''
        try:
            xyz = self.rowland.xyz_from_radiusangle(r, phi, x_range)
        except ValueError as e:
            if 'f(a) and f(b) must have different signs' in str(e):
                raise ElementPlacementError('No intersection with Rowland torus in range {0}'.format(self.x_range))
            else:
                # Something else went wrong
                raise e
        return xyz

    def xyzwpos(self):
        radii = self.distribute_elements_on_radius()
        facet_pos = np.array([self.xyz_from_radiusangle(r, self.phi, self.x_range).flatten() for r in radii])
        return e2h(facet_pos, 1)


    def max_elements_on_radius(self, radius):
        '''Distribute elements on a radius.

        Parameters
        ----------
        radius : list of two floats
            inner and outer radius that should be covered by elements

        Returns
        -------
        n : int
            Number of elements needed to cover a given radius segment.
            Elements might reach beyond the radius limits if the difference between
            inner and outer radius is not an integer multiple of the element size.
        '''
        return int(np.ceil((radius[1] - radius[0]) / self.d_element))

    def distribute_elements_on_radius(self):
        '''Distributes elements as evenly as possible along a radius.

        .. note::
           Unlike `distribute_elements_on_arc`, this function will have
           elements reaching beyond the limits of the radius, if the distance
           between inner and outer radius is not an integer multiple of the
           element size.

        Returns
        -------
        radii : np.ndarray
            Radii of the element *center* positions.

        '''
        radii = []
        for i in range(len(self.radius) // 2):
            radiusbracket = self.radius[2 * i: 2 * i + 2]
            n = self.max_elements_on_radius(radiusbracket)
            radii.append(np.mean(radiusbracket) +
                         np.arange(- n / 2 + 0.5, n / 2 + 0.5) * self.d_element)
        return np.hstack(radii)


class GratingArrayStructure(LinearCCDArray):
    '''A collection of diffraction gratings on the Rowland torus.

    When a ``GratingArrayStructure`` (GAS) is initialized, it places elements
    in the space available on the Rowland circle, most commonly, this class is
    used to place grating facets.

    After generation, individual facet positions can be adjusted by hand by
    editing the attributes `elem_pos` or `elem_uncertainty`. See `Parallel` for
    details.

    After any of the `elem_pos`, `elem_uncertainty` or `uncertainty` is
    changed, `generate_elements` needs to be called to regenerate the facets on
    the GAS.

    Parameters
    ----------
    rowland : RowlandTorus
    d_element : float
        Size of the edge of a element, which is assumed to be flat and square.
        (``d_element`` can be larger than the actual size of the silicon
        membrane to accommodate a minimum thickness of the surrounding frame.)

    x_range: list of 2 floats
        Minimum and maximum of the x coordinate that is searched for an
        intersection with the torus. A ray can intersect a torus in up to four
        points. ``x_range`` specififes the range for the numerical search for
        the intersection point.

    radius : list of 2 floats
        Inner and outer radius of the GAS as measured in the yz-plane from the
        origin.

    phi : list of 2 floats
        Bounding angles for a segment covered by the GSA. :math:`\phi=0` is on
        the positive y axis. The segment fills the space from ``phi1`` to
        ``phi2`` in the usual mathematical way (counterclockwise).  Angles are
        given in radian. Note that ``phi[1] < phi[0]`` is possible if the
        segment crosses the y axis.

    Notes
    -----
    This class derives from `LinearCCDArray`, which is a 1D arrangement of
    elements.  `GratingArrayStructure` also picks radii, but places several
    elements at each radius.

    '''

    tangent_to_torus = False
    '''If ``True`` the default orientation (before applying blaze, uncertainties etc.) of elements is
    such that they are tangents to the torus in the center of the element.
    If ``False`` they are perpendicular to perfectly focussed rays.
    '''

    id_col = 'facet'

    def __init__(self, rowland, d_element, x_range, radius, phi=[0., 2*np.pi],
                 parallel_spec=np.array([0., 1., 0., 0.]), **kwargs):
        if np.min(radius) < 0:
            raise ValueError('Radius must be positive.')
        kwargs['parallel_spec'] = parallel_spec
        kwargs['pos_spec'] = self.xyzwpos

        super(GratingArrayStructure, self).__init__(rowland, d_element, x_range, radius, phi, **kwargs)

    def calc_ideal_center(self):
        '''Position of the center of the GSA, assuming placement on the Rowland circle.'''
        a = (self.phi[0] + anglediff(self.phi) / 2 ) % (2. * np.pi)
        r = sum(self.radius) / 2
        return self.rowland.xyz_from_radiusangle(r, a, self.x_range).flatten()

    def max_elements_on_arc(self, radius):
        '''Calculate maximal number of elements that can be placed at a certain radius.

        Parameters
        ----------
        radius : float
            Radius of circle where the centers of all elements will be placed.
        '''
        return radius * anglediff(self.phi) // self.d_element

    def distribute_elements_on_arc(self, radius):
        '''Distribute elements on an arc.

        The elements are distributed as evenly as possible over the arc.

        .. note::

          Contrary to `distribute_elements_on_radius`, elements never stretch
          beyond the limits set by the ``phi`` parameter of the GAS. If an arc
          segment is not wide enough to accommodate at least a single element,
          it will go empty.

        Parameters
        ----------
        radius : float
            radius of arc where the elements are to be distributed.

        Returns
        -------
        centerangles : array
            The phi angles for centers of the elements at ``radius``.

        '''
        # arc is most crowded on inner radius
        n = self.max_elements_on_arc(radius - self.d_element / 2)
        element_angle = self.d_element / (2. * np.pi * radius)
        # thickness of space between elements, distributed equally
        d_between = (anglediff(self.phi) - n * element_angle) / (n + 1)
        centerangles = d_between + 0.5 * element_angle + np.arange(n) * (d_between + element_angle)
        return (self.phi[0] + centerangles) % (2. * np.pi)

    def getelem_xyzw(self):
        pos = []
        radii = self.distribute_elements_on_radius()
        for r in radii:
            angles = self.distribute_elements_on_arc(r)
            for a in angles:
                pos.append(self.rowland.xyz_from_radiusangle(r, a, self.x_range).flatten())
        return e2h(np.array(pos), 1)

    def calculate_elempos(self):
        '''Calculate the position of elements based on some algorithm.

        Returns
        -------
        pos4d : list of arrays
            List of affine transformations that bring an optical element
            centered on the origin of the coordinate system with the active
            plane in the yz-plane to the required facet position on the Rowland
            torus.
        '''
        pos4d = []

        xyzw = self.getelem_xyzw()
        normals = self.get_spec('normal_spec', xyzw)
        parallels = self.get_spec('parallel_spec', xyzw, normals)

        for i in range(xyzw.shape[0]):

            # Find the rotation between [1, 0, 0] and the new normal
            # Keep grooves (along e_y) parallel to e_y
            rot_mat = ex2vec_fix(h2e(normals[i, :]), h2e(parallels[i, :]))

            pos4d.append(transforms3d.affines.compose(h2e(xyzw[i, :]), rot_mat, np.ones(3)))
        return pos4d


class RectangularGrid(ParallelCalculated, OpticalElement):
    '''A collection of diffraction gratings on the Rowland torus.

    This class is similar to ``marxs.design.rowland.GratingArrayStructure`` but
    instead of placing elements on concentric circles, they are placed to fill
    a rectangular area.

    When initialized, it places elements in the space available on the
    Rowland circle, most commonly, this class is used to place grating facets.

    After generation, individual facet positions can be adjusted by hand by
    editing the attributes `elem_pos` or `elem_uncertainty`. See
    `marxs.simulation.Parallel` for details.

    After any of the `elem_pos`, `elem_uncertainty` or
    `uncertainty` is changed, `generate_elements` needs to be
    called to regenerate the facets on the GAS.

    Parameters
    ----------
    rowland : RowlandTorus
    d_element : float
        Size of the edge of a element, which is assumed to be flat and square.
        (``d_element`` can be larger than the actual size of the silicon
        membrane to accommodate a minimum thickness of the surrounding frame.)
    x_range: list of 2 floats
        Minimum and maximum of the x coordinate that is searched for an
        intersection with the torus. A ray can intersect a torus in up to four
        points. ``x_range`` specififes the range for the numerical search for
        the intersection point.
    y_range, z_range: lost of two floats
        limits of the rectangular area where gratings are placed.

    '''

    id_col = 'facet'

    def __init__(self, **kwargs):
        self.x_range = kwargs.pop('x_range')
        self.y_range = kwargs.pop('y_range')
        self.z_range = kwargs.pop('z_range')
        self.rowland = kwargs.pop('rowland')
        self.d_element = kwargs.pop('d_element')
        kwargs['pos_spec'] = self.elempos
        if 'normal_spec' not in kwargs.keys():
            kwargs['normal_spec'] = np.array([0., 0., 0., 1.])
        if 'parallel_spec' not in kwargs.keys():
            kwargs['parallel_spec'] = np.array([0., 1., 0., 0.])

        super(RectangularGrid, self).__init__(**kwargs)

    def elempos(self):

        n_y = int(np.ceil((self.y_range[1] - self.y_range[0]) / self.d_element))
        n_z = int(np.ceil((self.z_range[1] - self.z_range[0]) / self.d_element))

        # n_y and n_z are rounded up, so they cover a slighty larger range than y/z_range
        width_y = n_y * self.d_element
        width_z = n_z * self.d_element

        ypos = np.arange(0.5 * (self.y_range[0] - width_y + self.y_range[1] + self.d_element), self.y_range[1], self.d_element)
        zpos = np.arange(0.5 * (self.z_range[0] - width_z + self.z_range[1] + self.d_element), self.z_range[1], self.d_element)
        ypos, zpos = np.meshgrid(ypos, zpos)

        xpos = []
        for y, z in zip(ypos.flatten(), zpos.flatten()):
            xpos.append(self.rowland.solve_quartic(y=y, z=z, interval=self.x_range))

        return np.vstack([np.array(xpos), ypos.flatten(), zpos.flatten(), np.ones_like(xpos)]).T

    def calculate_elempos(self):
        '''Calculate the position of elements based on some algorithm.

        Returns
        -------
        pos4d : list of arrays
            List of affine transformations that bring an optical element centered
            on the origin of the coordinate system with the active plane in the
            yz-plane to the required facet position on the Rowland torus.
        '''
        pos4d = []

        xyzw = self.elempos()
        normals = self.get_spec('normal_spec', xyzw)
        parallels = self.get_spec('parallel_spec', xyzw, normals)

        for i in range(xyzw.shape[0]):

            # Find the rotation between [1, 0, 0] and the new normal
            # Keep grooves (along e_y) parallel to e_y
            rot_mat = ex2vec_fix(h2e(normals[i, :]), h2e(parallels[i, :]))

            pos4d.append(transforms3d.affines.compose(h2e(xyzw[i, :]), rot_mat, np.ones(3)))
        return pos4d



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest

from ..utils import (plane_with_hole, combine_disjoint_triangulations,
                     get_color, color_tuple_to_hex,
                     MARXSVisualizationWarning,
                     DisplayDict)
from ..mayavi import plot_object

def test_hole_round():
    '''Ensure that plane around the inner hole closes properly at the last point.'''
    outer = np.array([[-1, -1, 1, 1], [-1, 1, 1, -1], [0,0,0,0]]).T

    # Try out a rectangle and a circle - the two cases we are most likely to need.
    for n in [3, 4, 90, 360]:
        inner  = np.zeros((n, 3))
        phi = np.linspace(0, 2 * np.pi, n, endpoint=False)
        inner[:, 0] = np.sin(phi)
        inner[:, 1] = np.cos(phi)
        xyz, triangles = plane_with_hole(outer, inner)
        assert triangles[0][1] == triangles[-1][2]
        tri = np.array(triangles)
        # Technically, the order of the points in each triangle
        # is irrelevant, but the current implementation does it this way
        # and that's an easy way to check.
        # Check first point is always on outer rim
        assert set(tri[:, 0]) == set([0,1,2,3])
        # Check all points turn up in the middle
        assert set(tri[:, 1]) == set(np.arange(4 + n))
        # Check last point is always on inner rim
        assert set(tri[:, 2]) == set(np.arange(4, 4 + n))

def test_stack_triangulations():
    xyz = np.array([[-1. , -1. ,  0. ],
                    [-1. ,  1. ,  0. ],
                    [ 1. ,  1. ,  0. ],
                    [ 1. , -1. ,  0. ]])
    triangles = np.array([[0, 1, 2], [0, 2, 3]])
    xyz_out, tri_out = combine_disjoint_triangulations([xyz, xyz + 5.3],
                                                       [triangles, triangles])
    assert np.all(xyz_out[:4, :] == xyz)
    assert np.allclose(xyz_out[4:, :], xyz + 5.3)
    assert np.all(tri_out[:2, :] == triangles)
    assert np.all(tri_out[2:, :] == triangles + 4)

def test_color_roundtrip():
    '''Test that the different color convertes are consistent.'''
    colortup = get_color({'color': 'white'})
    assert colortup == (1.0, 1.0, 1.0)
    hexstr = color_tuple_to_hex(colortup)
    assert hexstr == '0xffffff'
    assert get_color({'color': '#ffffff'}) == (1.0, 1.0, 1.0)
    # repeat with a second color that has different values for rgb
    colortup = get_color({'color': '#ff013a'})
    # Note that matplotlib expects hex strings with '#' while python uses '0x'
    assert color_tuple_to_hex(colortup) == '0xff013a'

    # Int input
    assert color_tuple_to_hex((255, 255, 255)) == '0xffffff'

def test_color_hex_pad():
    '''Regression test: We want to color hexstring with leading zeroths'''
    assert color_tuple_to_hex((0., 1., .5)) == '0x00ff7f'

def test_color_to_hex_bad_input():
    with pytest.raises(ValueError) as e:
        out = color_tuple_to_hex('white')
    assert 'Input tuple must be all' in str(e.value)

    with pytest.raises(ValueError) as e:
        out = color_tuple_to_hex((-1, 0, 234))
    assert 'Int values in color tuple' in str(e.value)

    with pytest.raises(ValueError) as e:
        out = color_tuple_to_hex((1., 3., 0.3))
    assert 'Float values in color tuple' in str(e.value)

def test_no_display_warnings():
    '''Check that warnings are emitted without for classes, functions, and others'''
    class NoDisplay(object):
        pass

    with pytest.warns(MARXSVisualizationWarning) as record:
        plot_object(NoDisplay())

    # check that only one warning was raised
    assert len(record) == 1
    # check that the message matches
    assert 'No display dictionary found.' in record[0].message.args[0]

    def f():
        pass

    with pytest.warns(MARXSVisualizationWarning) as record:
        plot_object(f)

    # check that only one warning was raised
    assert len(record) == 1
    # check that the message matches
    assert 'No display dictionary found.' in record[0].message.args[0]


    with pytest.warns(MARXSVisualizationWarning) as record:
        plot_object(range)

    assert len(record) == 1
    assert 'No display dictionary found' in record[0].message.args[0]

def test_warning_unknownshape():
    '''Test warning (and no crash) for plotting stuff of unknown shape.'''
    class Display():
        display = {'shape': 'cilinder'}

    with pytest.warns(MARXSVisualizationWarning) as record:
        plot_object(Display())

    # check that only one warning was raised
    assert len(record) == 1
    # check that the message matches
    assert 'No function to plot cilinder.' in record[0].message.args[0]

def test_warning_noshapeset():
    '''Test warning (and no crash) for plotting stuff of unknown shape.'''
    class Display():
        display = {'form': 'cilinder'}

    with pytest.warns(MARXSVisualizationWarning) as record:
        plot_object(Display())

    # check that only one warning was raised
    assert len(record) == 1
    # check that the message matches
    assert '"shape" not set in display dict.' in record[0].message.args[0]

def test_DiplayDict():
    '''Simple mock-up using a DisplayDict variable.'''
    class A(object):
        pass

    a = A()
    b = A()
    b.geometry = A()
    b.geometry.value = 5

    a.display = DisplayDict(a, r=7)
    b.display = DisplayDict(b, value_1='w')

    assert a.display['r'] == 7
    assert b.display['value_1'] == 'w'
    assert b.display['value'] == 5
    # value is set in display and in geometry
    b.display['value'] = 6
    assert b.display['value'] == 6
    assert b.geometry.value == 5
    # values added to geometry later
    b.geometry.value2 = 'q'
    assert b.display['value2'] == 'q'
    # Make sure there is the right error only when used on an object without geomtry
    with pytest.raises(KeyError) as e:
        temp = a.display['q']
    # Make sure there is the right error if something is not found
    with pytest.raises(KeyError) as e:
        temp = b.display['1']

    # get works, too
    assert b.display.get('value') == 6
    assert a.display.get('qwer') is None
    assert a.display.get('asdf', 5) == 5



# Licensed under GPL version 3 - see LICENSE.rst
'''Add additional scattering to processed photons.

Classes in this file add additional scattering in a statistical sense.

The classes in this module do not trace the photon to a specific location, they
just add the scatter at the point of the last interaction.  For example,
reflection from a (flat) mirror is implemented as a perfect reflection, but in
practice there is some roughness to the mirror that adds a small Gaussian blur
to the reflection. To represent this, the point of origin of the ray remains
unchanged, but a small random change is added to the direction vector.

'''
import numpy as np
from warnings import warn
import astropy.units as u

from ..math.utils import e2h, h2e, norm_vector
from ..math.rotations import axangle2mat
from ..math.polarization import parallel_transport
from .base import FlatOpticalElement

class RadialMirrorScatter(FlatOpticalElement):
    '''Add scatter to any sort of radial mirror.

    Scatter is added in the plane of reflection, which is defined here
    as the plane which contains (i) the current direction the ray and (ii) the
    vector connecting the center of the `RadialMirrorScatter` element and the
    point of last interaction for the ray.
    Scatter can also be added perpendicular to the plane-of-reflection.

    Parameters
    ----------
    inplanescatter : `astropy.Quantity`
        sigma of Gaussian for in-plane scatter
    perpplanescatter : `astropy.Quantity`
        sigma of Gaussian for scatter perpendicular to the plane of reflection
        (default = 0)
    '''
    def __init__(self, **kwargs):
        self.inplanescatter = kwargs.pop('inplanescatter').to(u.rad).value
        self.perpplanescatter = kwargs.pop('perpplanescatter', 0. * u.rad).to(u.rad).value
        super(RadialMirrorScatter, self).__init__(**kwargs)

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        n = intersect.sum()
        center = self.pos4d[:-1, -1]
        radial = h2e(photons['pos'][intersect].data) - center
        perpplane = np.cross(h2e(photons['dir'][intersect].data), radial)

        # np.random.normal does not work with scale=0
        # so special case that here.
        if self.inplanescatter != 0:
            inplaneangle = np.random.normal(loc=0., scale=self.inplanescatter, size=n)
            rot = axangle2mat(perpplane, inplaneangle)
            outdir = e2h(np.einsum('...ij,...i->...j', rot, h2e(photons['dir'][intersect])), 0)
        else:
            inplaneangle = np.zeros(n)
            outdir = photons['dir'][intersect]

        if self.perpplanescatter !=0:
            perpangle = np.random.normal(loc=0., scale=self.perpplanescatter, size=n)
            rot = axangle2mat(radial, perpangle)
            outdir = e2h(np.einsum('...ij,...i->...j', rot, h2e(outdir)), 0)
        else:
            perpangle = np.zeros_like(inplaneangle)

        pol = parallel_transport(photons['dir'].data[intersect, :], outdir,
                                 photons['polarization'].data[intersect, :])

        return {'dir': outdir, 'polarization': pol,
                'inplanescatter': inplaneangle, 'perpplanescatter': perpangle}


class RandomGaussianScatter(FlatOpticalElement):
    '''Add scatter to any sort of radial mirror.

    This element scatters rays by a small angle, drawn from a Gaussian
    distribution. The direction of the scatter is random.

    Parameters
    ----------
    scatter : `astropy.Quantitiy` or callable
        This this is a number, scattering angles will be drawn from a Gaussian
        with the given sigma. For a variable scatter, this can be a
        function with the following call signature: ``angle = func(photons,
        intersect, interpos, intercoos)``. The function should return an
        `astropy.Quantity` array, containing one angle for each intersecting
        photon. A function passed in for this parameter can makes the
        scattering time, location, or energy-dependent.
    '''
    scattername = 'scatter'

    def __init__(self, **kwargs):
        if 'scatter' in kwargs:
            if hasattr(self, 'scatter'):
                warn('Overriding class level "scatter" definition.')
            self.scatter = kwargs.pop('scatter') # in rad
        else:
            if not hasattr(self, 'scatter'):
                raise ValueError('Keyword "scatter" missing.')
        super().__init__(**kwargs)

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        n = intersect.sum()
        # np.random.normal does not work with scale=0
        # so special case that here.
        with u.set_enabled_equivalencies(u.dimensionless_angles()):
            scatterzero = self.scatter == 0
        if scatterzero:
            angle = np.zeros(n)
            out = {}
        else:
            pdir = norm_vector(h2e(photons['dir'][intersect].data))
            # Now, find a direction that is perpendicular to the photon direction
            # Any perpendicular direction will do
            # Start by making a set of vectors that at least are not parallel
            # to the photon direction
            guessvec = np.zeros_like(pdir)
            ind = np.abs(pdir[:, 0]) < 0.99999
            guessvec[ind, 0] = 1
            guessvec[~ind, 1] = 1
            perpvec = np.cross(pdir, guessvec)
            if callable(self.scatter):
                angle = self.scatter(photons, intersect,
                                     interpos, intercoos).to(u.rad).value
            else:
                angle = np.random.normal(loc=0.,
                                         scale=self.scatter.to(u.rad).value,
                                         size=n)
            rot = axangle2mat(perpvec, angle)
            outdir = np.einsum('...ij,...i->...j', rot, pdir)
            # Now rotate result by up to 2 pi to randomize direction
            angle2 = np.random.uniform(size=n) * 2 * np.pi
            rot = axangle2mat(pdir, angle2)
            outdir = e2h(np.einsum('...ij,...i->...j', rot, outdir), 0)
            pol = parallel_transport(photons['dir'].data[intersect, :], outdir,
                                     photons['polarization'].data[intersect, :])
            out = {'dir': outdir, 'polarization': pol}
        if self.scattername is not None:
            out[self.scattername] = angle
        return out



# Licensed under GPL version 3 - see LICENSE.rst
'''This module collects routines that deal with lines in Pluecker coordiantes incl. the
construction of the Pluecker coordinates from e.g. two points in space or a direction and
a point.
'''
import numpy as np

def angle_line_plane(p_line, h_plane):
    '''Compute angle between normal to plane and line

    Parameters
    ----------
    p_line : array-like
        Line(s) in Pluecker coordinates
    h_plane : array-like
        Plane in homogeneous coordinates

    Returns
    -------
    angle : array-like
    '''
    l = p_line[:3]
    p = h_plane[:3]
    return np.arccos(np.dot(l, p) / (np.linalg.norm(l) * np.linalg.norm(p)))


#  @ L = {U:V}, with 3-tuples U and V, with U.V = 0, and with U non-null.

def e_pointpoint2line(e_p1, e_p2):
    '''Return a line in Pluecker coordinates connecting two points.

    p1, p2 are distinct point on line, direction is p1 -> p2

    Parameters
    ----------
    e_p1, e_p2: array_like
        Array of Eucledian positions

    Returns
    -------
    p_line : array-like
        Lines in Pluecker coordinates
    '''
    p_line = np.empty(e_p1.shape[:-1] + (6,) )
    p_line[..., :3] = e_p2 - e_p1
    p_line[..., 3:] = np.cross(e_p2, e_p1)
    return p_line


def dir_point2line(e_dir, e_pos):
    '''Return a line in Pluecker coordinates connecting two points.

    p1, p2 are distinct point on line, direction is p1 -> p2

    Parameters
    ----------
    e_dir: array_like
        Array of Euclidean direction vectors
    e_point : array-like
        Array of points in eukledian coordinates

    Returns
    -------
    p_line : array-like
        Lines in Pluecker coordinates
    '''
    p_line = np.empty(e_dir.shape[:-1] + (6,) )
    p_line[..., :3] = e_dir
    p_line[..., 3:] = np.cross(e_dir, e_pos)
    return p_line


def intersect_line_plane(p_line, h_plane):
    '''Intersect a line in Pluecker coordinates with a plane in homogeneous coordinates.

    This function returns the position of the intersection.
    If the line is parallel (but not identical) to the plane that intersection
    happens "at infinity" and thus the last component of the intersection
    coordinate is 0. If the line is in the plane, then all components of the
    return vector are 0.

    Parameters
    ----------
    p_line : np.array (last dimension has 6 elements)
        Pluecker coordinates of a line or an array of lines
    h_plane : np.array with ``shape=(4, )``
        Homogeneous coordinates of a plane.

    Returns
    -------
    h_point : np.array
        Homogeneous coordinates of the intersection points of line and plane.
        Has the same shape as ``p_line``, but the last dimension has only 4
        elements.
    '''
    h_point = np.empty(p_line.shape[:-1] + (4,)  )
    h_point[..., :3] = np.cross(p_line[..., 3:], h_plane[:3]) - p_line[..., :3] * h_plane[3]
    h_point[..., 3] = np.dot(p_line[..., :3], h_plane[:3])
    return h_point


def point_dir2plane(h_point, h_dir):
    '''Return a plane in homogeneous coordinates

    Parameters
    ----------
    h_point : array-like
        Point in plane in homogeneous coordinates
    h_dir: array_like
        Normal vector of plane in homogeneous coordinates

    Returns
    -------
    p_line : array-like
        Lines in Pluecker coordinates
    '''
    h_plane = np.empty(4)
    h_dir_n = h_dir / np.linalg.norm(h_dir)
    h_plane[:3] = h_dir_n[:3]
    h_plane[3] = - np.dot(h_point, h_dir_n) / h_point[3]
    return h_plane


#   @ L = {U:UxQ}, for U the direction of L and Q a point on L.
#   @ L = {qP-pQ:PxQ}, for (P:p) and (Q:q) distinct homogeneous points on L.
#   @ L = {ExF:fE-eF}, for [E:e] and [F:f] distinct planes containing L.

#   @ {U1:V1} =? s{U2:V2} tests if L1 = {U1:V1} equals L2 = {U2:V2}.
#   @ s > 0 if L1 and L2 have same orientation.

#   @ (V.V)/(U.U) is the minimum squared distance of L from the origin.
#   @ (VxU:U.U) is the point of L closest to the origin.
#   @ [UxV:V.V] is the plane through L perpendicular to its origin plane, for
#         non-null V.

#   @ (VxN-Un:U.N) is the point where L intersects plane [N:n] not parallel to L.
#   @ [UxP-Vw:V.P] is the plane containing L and point (P:w) not on L.
#   @ [UxN:V.N] is the plane containing L and direction N not parallel to L.

#   Let N, N1, N2 be unit vectors along the coordinate axes, with U.N non-zero.

#   @ (VxN:U.N) is a point on L if N is not perpendicular to U.
#   @ U and this point both satisfy a plane equation [E:e] if the plane
#         contains L.
#   @ Represent L as U and this point to transform by non-perspective
#         homogeneous matrix.
#   @ Represent L as two points to transform by perspective homogeneous matrix.

#   @ [UxN1:V.N1] and [UxN2:V.N2] are distinct planes containing L.
#   @ P satisfies both these plane equations if L contains P.

#   @ Pnt(t) = (VxU+tU:U.U) parameterizes points on L.
#   @ Pln(t) = (1-t^2)[UxN1:V.N1]+2t[UxN2:V.N2] parameterizes planes through L.

#   @ U1.V2 + U2.V1 =? 0 tests if L1 = {U1:V1} and L2 = {U2:V2} are coplanar
#         (intersect).
#   @ Sum positive if right-handed screw takes one into the other; negative
#         if left-handed.
#   @ U1xU2 =? 0 tests if lines are parallel.
#   Let N be a unit vector along a coordinate axis, with (U1xU2).N non-zero.
#   @ ((V1.N)U2-(V2.N)U1-(V1.U2)N:(U1xU2).N) is the point of intersection, if
#         any.

#   @ [U1xU2:V1.U2] is the common plane for non-parallel lines.
#   Let N, N1, N2 be unit vectors along the coordinate axes, with U1.N non-zero.
#   @ [(U1.N)V2-(U2.N)V1:(V1xV2).N] is the common plane for parallel distinct
#         lines.
#   @ [U1xN1:V1.N1] is the common plane for equal lines through origin.

# Here are two related tricks, a lagniappe, as they say in New Orleans.

#   Let P be the point (x,y,z).
#   @ [(-1,0,0):x] and [(0,-1,0):y] and [(0,0,-1):z] are independent planes
#         through P.

#   Let [E:e] be a plane and N, N1, and N2 unit coordinate axis vectors with
#         E.N non-null.
#   @ Point (-eN:E.N) and distinct direction vectors ExN1 and ExN2 lie in the
#         plane.



# Licensed under GPL version 3 - see LICENSE.rst

import numpy as np

from .fitsheaders import complete_header
from ...simulator import Sequence


class Chandra(Sequence):
    '''Main class representing the Chandra X-ray observatory


    '''
    def process_photons(self, photons):
        photons.meta['MISSION'] = ('AXAF', 'Mission')
        photons.meta['TELESCOP'] = ('CHANDRA', 'Telescope')
        return super(Chandra, self).process_photons(photons)

    def write_evt(self, photons, filename):
        '''Write a Chandra event level 1 file.

        As opposed to directly saving the photon list, this adds some Chandra
        specific meta data.

        Parameters
        ----------
        photons :  `astropy.table.Table` or `astropy.table.Row`
            Table with photon properties. Some meta data from the header of this table
            is required (e.g. the length of the observation).
        filename : string
            Path and file name where the file is saved.
        '''
        photons.meta['EXTNAME'] = 'EVENTS'
        # rename RA, DEC columns - otherwise CIAO tasks will be confused.
        photons.rename_column('ra', 'marxs_ra')
        photons.rename_column('dec', 'marxs_dec')
        complete_header(photons.meta, photons, 'EVT1', ['OGIP', 'EVENTS', 'ALL'])
        photons.write(filename, format='fits')
        # add_GTIs(filename)



# Licensed under GPL version 3 - see LICENSE.rst
from collections import OrderedDict
import inspect
import warnings
from copy import deepcopy

import numpy as np
from transforms3d import affines
from astropy.table import Column

from ..visualization.utils import DisplayDict

class GeometryError(Exception):
    pass


class DocMeta(type):
    '''Metaclass to inherit docstrings when reqired.

    When a derived class overwrites a method that was already defined in its
    base class, the new method usually has the same purpose as the original
    method and often uses the same parameters, too, although the implementation
    differs slightly.  In this case, it should have the same docstring, too.
    This metaclass will look for methods that are undocumented and add the
    docstring of the appropriate parent method to them.

    '''
    def __new__(mcs, name, bases, dict):
        # make a class here with the same method resolution order
        # but no attributes of its own (so that we can make it here without an
        # infinite loop, because making it will also go through this metaclass)
        if name == 'temporaryclass':
            return type.__new__(mcs, name, bases, dict)
        temp = type('temporaryclass', bases, {})
        mro = inspect.getmro(temp)

        for k in dict:
            # for items that have a docstring (i.e. methods) that is empty
            if hasattr(dict[k], '__doc__') and dict[k].__doc__ is None:
                # Some __doc__ strings cannot be rewitten in Python2, e.g. doc of a type
                try:
                    for b in mro:
                        # look if this b defines the method
                        # (it might not in case of multiple inheritance)
                        if hasattr(b, k) and (getattr(b, k).__doc__ is not None):
                            # copy docstring if super method has one
                            dict[k].__doc__ = getattr(b, k).__doc__
                            break
                except AttributeError:
                    pass

        return type.__new__(mcs, name, bases, dict)


class MarxsElement(metaclass=DocMeta):
    '''Base class for all elements in a MARXS simulation.

    This includes elements that actually change photon properties such as grating and
    mirrors, but also abstract concepts that do not have a direct hardware
    representation such as a "Rowland Torus".
    '''

    display = {'shape': 'None'}
    'Dictionary for display specifications, e.g. color'

    def __init__(self, **kwargs):
        '''Define a new MARXS element.'''
        if 'name' in kwargs:
            self.name = kwargs.pop('name')
        else:
            self.name = self.__class__

        if len(kwargs) > 0:
            raise ValueError('Initialization arguments {0} not understood'.format(', '.join(kwargs.keys())))
        self.display = DisplayDict(self, deepcopy(self.display))

    def describe(self):
        return OrderedDict(element=self.name)

class SimulationSequenceElement(MarxsElement):
    '''Base class for all elements in a simulation sequence that processes photons.'''

    output_columns = []
    '''This is a list of strings that names the output properties.

    This gives the names of the output properties from this optical element.
    `process_photon` or `process_photons` are responsible for calculating the values of these
    properties. For example, for a mirror of nested shells one might set
    ``output_columns = ['mirror_shell']`` to pass the information on which shell the interaction
    took place to the user.

    The following properties are always included in the output and do not need to be listed here:

        dir : `numpy.ndarray`
            4-d direction vector of ray in homogeneous coordinates
        pos : `numpy.ndarray`
            4-d position of last interaction pf the photons with any optical
            element in homogeneous coordinates. Together with ``dir`` this
            determines the equation of the ray.
        energy : float
            Photon energy in keV.
        polarization : float
            Polarization angle of the photons.
        probability : float
            Probability that the photon continues. Set to 0 if the photon is
            absorbed, to 1 if it passes the optical element and to number
            between 0 and 1 to express a probability that the photons passes.
    '''

    id_col = None
    '''String that names an id column for output.

    Set this to a string to add an automatic numbering to the output. This is
    especially useful if there are several identical optical components that
    are used in parallel, e.g. there are four identical CCDs. Setting ``id_col
    = "CCD_ID"`` and passing an ``id_num=1, 2, 3, 4`` keyword respectively to
    each CCD will add a column ``CCD_ID`` with a value of 1,2,3, or 4 for each
    photon hitting one of those CCDs.

    Currently, this will not work with all optical elements.

    '''

    def __init__(self, **kwargs):
        self.id_num = kwargs.pop('id_num', -9)
        # We want to use id_col as a class attribute, but overwrite it if given as a kwarg
        if 'id_col' in kwargs:
            self.id_col = kwargs.pop('id_col')
        super(SimulationSequenceElement, self).__init__(**kwargs)

    def add_output_cols(self, photons, colnames=[]):
        '''Add output columns of the correct format (currently: float) to the photon array.

        This function takes the column names that are added to ``photons`` from
        several sources:

        - `id_col` (if not ``None``)
        - `output_columns`
        - the ``colnames`` parameter.

        Parameters
        ----------
        photons : `astropy.table.Table`
            Table columns are added to.
        colnames : list of strings
            Column names to be added; in addition several object properties can
            be used to set the column names, see description above.
        '''
        temp = np.empty(len(photons))
        temp[:] = np.nan
        for n in self.output_columns + colnames:
            if n not in photons.colnames:
                photons.add_column(Column(name=n, data=temp))

        if self.id_col is not None:
            if self.id_col not in photons.colnames:
                photons.add_column(Column(name=self.id_col, data=-np.ones(len(photons))))

    def __call__(self, photons, *args, **kwargs):
        return self.process_photons(photons, *args, **kwargs)


def _parse_position_keywords(kwargs):
    '''Parse keywords to define position.

    If ``pos4d`` is given, use that, otherwise look for ``position``, ``zoom``
    and ``orientation``.

    Parameters
    ----------
    pos4d : 4x4 array
        Transformation to bring an element from the default position (see description of
        individual elements) to a certain point in the space of homogeneous coordinates.
    position : 3-d vector in real space
        Measured from the origin of the spacecraft coordinate system.
    orientation : Rotation matrix or ``None``
        Relative orientation of the base vectors of this optical
        element relative to the orientation of the coordinate system
        of the spacecraft. The default is no rotation (i.e. the axes of both
        coordinate systems are parallel).
    zoom : float or 3-d vector
        Scale the size of an optical element in each dimension by ``zoom``. If ``zoom`` is a scalar,
        the same scale is applied in all dimesions. This only affects the oter dimensions of
        the optical element, not internal scales like the pixel size or grating constant (if
        defined for the optical element in question).
    '''
    pos4d = kwargs.pop('pos4d', None)
    if pos4d is None:
        position = kwargs.pop('position', np.zeros(3))
        orientation = kwargs.pop('orientation', np.eye(3))
        zoom = kwargs.pop('zoom', 1.)
        if np.isscalar(zoom):
            zoom = np.ones(3) * zoom
        if not len(zoom) == 3:
            raise ValueError('zoom must have three elements for x,y,z or be a scalar (global zoom).')
        if np.any(np.array(zoom) <= 0):
            raise ValueError('All values in zoom must be positive-definite to keep pos4d matrix valid. Specify zero-thickness elements with display properties.')
        pos4d = affines.compose(position, orientation, zoom)
    else:
        if ('position' in kwargs) or ('orientation' in kwargs) or ('zoom' in kwargs):
            raise ValueError('If pos4d is specificed, the following keywords cannot be given at the same time: position, orientation, zoom.')

    if np.linalg.det(pos4d) == 0:
        raise ValueError('pos4d matrix is invalid (determinant is 0).')
    return pos4d



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import transforms3d
import pytest
import astropy.units as u

from ..labSource import LabPointSourceCone


def test_labsource_eukledian():
    with pytest.raises(ValueError) as err:
        source = LabPointSourceCone(position=[1, 2, 3, 1.])
    assert 'in Eukledian coordinates' in str(err.value)


def test_photon_generation():
    '''This tests the lab point source. It checks that the starting points
    are all at the sources position.
    '''
    pos = [1., 1., 1.]
    rate = 10 / u.s
    source = LabPointSourceCone(position=pos, flux=rate, energy=5. * u.keV)

    photons = source.generate_photons(1. * u.s)
    assert np.all(photons['pos'] == np.ones([10, 4]))


def test_directions_range_cone():
    '''This tests that all of the photons in the returned table have directions within the given parameters.
    - Ensures that all direciton vectors are normed
    - Ensures that all direction vectors are within specified range
    '''

    # parameters
    pos = [5 * np.random.random(), 5 * np.random.random(), 5 * np.random.random()]
    rate = 10 * np.random.random() / u.s
    direction = [5* np.random.random(),5* np.random.random(),5* np.random.random()]
    delta = ((np.pi / 2) * (np.random.random() * 2 - 1))

    # run simulation
    source = LabPointSourceCone(position=pos, half_opening=delta, flux=rate,
                                energy=5. * u.keV, direction = direction)
    photons = source.generate_photons(1. * u.s)


    # norm direction
    direction = direction / np.sqrt(np.dot(direction, direction))

    # if the photon direction has the correct magnitude, then it will be on the unit sphere. All that is left to assert, then, is that it is sufficiently close to the direction vector.
    # take any orthogonal axis to direction. By symmetry it will be the same for any orthogonal axis.
    if (direction[0] == 0) and (direction[1]== 0):
        if (direction[2] == 0):
            #direction is zero vector
            raise ValueError('zero vector')
        else:
            axis = [0, 1, 0]
    else:
        axis = [-direction[1], direction[0], 0]

    # now we find a direction vector that is rotated the maximum distance from the beam direction
    rotationMatrix = transforms3d.axangles.axangle2mat(axis, delta)
    farVector = np.dot(rotationMatrix, direction)
    displacement = farVector - np.array(direction)
    maxDistanceSquared = np.dot(displacement, displacement)

    # Assert (correct magnitude) and (sufficiently close)

    # Prepares assessment of vector length
    directionMatrix = np.array(photons['dir']) # 2D array of photon directions
    dotProducts = np.sum((directionMatrix * directionMatrix), axis = 1) # the magnitudes^2 of each vector in an array
    correctSizes = [ (abs(1 - i) < 0.001) for i in dotProducts]

    # Prepares assessment of direction
    direction = np.array( np.append(direction, [0], axis = 0) )
    photonDisplacements = directionMatrix - direction
    displacementDistancesSquared = np.sum((photonDisplacements * photonDisplacements), axis = 1)
    withinRange = [(maxDistanceSquared > i) for i in displacementDistancesSquared]

    assert all(correctSizes) and all(withinRange)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table
from astropy.coordinates import SkyCoord
import astropy.units as u
from scipy.stats import pearsonr
import pytest

from ...source import FixedPointing, JitterPointing, PointSource


def test_reference_coordiante_system():
    '''Usually, rays that are on-axis will come in along the x-axis.
    Test a simulations with a different coordinate system.'''
    s = PointSource(coords=SkyCoord(12., 34., unit='deg'))
    photons = s.generate_photons(5 * u.s)
    point_x = FixedPointing(coords=SkyCoord(12., 34., unit='deg'))
    p_x = point_x(photons.copy())
    assert np.allclose(p_x['dir'].data[:, 0], -1.)
    assert np.allclose(p_x['dir'].data[:, 1:], 0)

    xyz2zxy = np.array([[0., 0, 1, 0], [1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1]]).T
    point_z = FixedPointing(coords=SkyCoord(12., 34., unit='deg'), reference_transform=xyz2zxy)
    p_z = point_z(photons.copy())
    assert np.allclose(p_z['dir'].data[:, 2], -1.)
    assert np.allclose(p_z['dir'].data[:, :2], 0)

def test_polarization_direction():
    '''Test that a FixedPointing correctly assigns linear polarization vectors.'''
    s = PointSource(coords=SkyCoord(187.4, 0., unit='deg'))
    photons = s.generate_photons(5 * u.s)
    photons['polangle'] = (np.array([0., 90., 180., 270., 45.])) * u.deg.to(photons['polangle'].unit)
    point_x = FixedPointing(coords=SkyCoord(187.4, 0., unit='deg'))
    p_x = point_x(photons.copy())
    # For a simple polangle 0 = 180 and the direction does not matter.
    # However with a view to eliptical polarization in the future we want
    # to fix the absolute phase, too.
    assert np.allclose(p_x['polarization'].data[0, :], [0, 0, 1, 0])
    assert np.allclose(p_x['polarization'].data[1, :], [0, 1, 0, 0])
    assert np.allclose(p_x['polarization'].data[2, :], [0, 0, -1, 0])
    assert np.allclose(p_x['polarization'].data[3, :], [0, -1, 0, 0])
    assert np.allclose(p_x['polarization'].data[4, :], [0, 1/np.sqrt(2), 1/np.sqrt(2), 0])

    xyz2zxy = np.array([[0., 0, 1, 0], [1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1]]).T
    point_z = FixedPointing(coords=SkyCoord(187.4, 0., unit='deg'), reference_transform=xyz2zxy)
    p_z = point_z(photons.copy())
    assert np.allclose(p_z['polarization'].data[0, :], [0, 1, 0, 0])
    assert np.allclose(p_z['polarization'].data[1, :], [1, 0, 0, 0])
    assert np.allclose(p_z['polarization'].data[4, :], [1/np.sqrt(2), 1/np.sqrt(2), 0, 0])

    # Photons pointing east with the same RA will have parallel polarization vectors
    # This is true for all pointing directions.
    s = PointSource(coords=SkyCoord(22.5, 0., unit='deg'))
    photons = s.generate_photons(5 * u.s)
    photons['dec'] = [67., 23., 0., -45.454, -67.88]
    photons['polangle'] = (90. * u.deg).to(photons['polangle'].unit)
    point = FixedPointing(coords=SkyCoord(94.3, 23., unit='deg'))
    p = point(photons.copy())
    for i in range(1, len(p)):
        assert np.isclose(np.dot(p['polarization'][0], p['polarization'][i]), 1)

def test_jitter():
    '''test size and randomness of jitter'''
    n = 100000
    ra = np.random.rand(n) * 2 * np.pi
    # Note that my rays are not evenly distributed on the sky.
    # No need to be extra fancy here, it's probably even better for
    # a test to have more scrutiny around the poles.
    dec = (np.random.rand(n) * 2. - 1.) / 2. * np.pi
    time = np.arange(n)
    pol = np.ones_like(ra)
    prob = np.ones_like(ra)
    photons = Table([ra, dec, time, pol, prob],
                    names=['ra', 'dec', 'time', 'polangle', 'probability'])
    fixed = FixedPointing(coords = SkyCoord(25., -10., unit='deg'))
    jittered = JitterPointing(coords = SkyCoord(25., -10., unit='deg'),
                              jitter=1. * u.arcsec)
    p_fixed = fixed(photons.copy())
    p_jitter = jittered(photons)

    assert np.allclose(np.linalg.norm(p_fixed['dir'], axis=1), 1.)
    assert np.allclose(np.linalg.norm(p_jitter['dir'], axis=1), 1.)

    prod = np.sum(p_fixed['dir'] * p_jitter['dir'], axis=1)
    # sum can give values > 1 due to rounding errors
    # That would make arccos fail, so catch those here
    ind = prod > 1.
    if ind.sum() > 0:
        prod[ind] = 1.

    alpha = np.arccos(prod)
    # in this formula alpha will always be the abs(angle).
    # Flip some signs to recover input normal distribtution.
    alpha *= np.sign(np.random.rand(n)-0.5)
    # center?
    assert np.abs(np.mean(alpha)) * 3600. < 0.01
    # Is right size?
    assert np.std(np.rad2deg(alpha)) > (0.9 / 3600.)
    assert np.std(np.rad2deg(alpha)) < (1.1 / 3600.)
    # Does it affect y and z independently?
    coeff, p = pearsonr(p_fixed['dir'][:, 1] - p_jitter['dir'][:, 1],
                        p_fixed['dir'][:, 2] - p_jitter['dir'][:, 2])
    assert abs(coeff) < 0.01

@pytest.mark.parametrize('pointing', [FixedPointing, JitterPointing])
def test_polarization_perpendicular(pointing):
    '''Consistency: Polarization vector must always be perpendicular to dir.'''
    s = PointSource(coords=SkyCoord(0., 0., unit='deg'))
    photons = s.generate_photons(10 * u.s)
    photons['ra'] = np.random.uniform(0, 360., len(photons))
    # Exclude +-90 deg, because not handling poles well
    photons['dec'] = np.random.uniform(-89.9, 89.9, len(photons))
    photons['polangle'] = np.random.uniform(0, 360., len(photons))
    point_x = FixedPointing(coords=SkyCoord(187.4, 0., unit='deg'))
    p_x = point_x(photons.copy())
    assert np.allclose(np.einsum('ij,ij->i', p_x['dir'].data,
                                 p_x['polarization'].data), 0)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import transforms3d
from astropy.table import Table
from astropy.coordinates import SkyCoord
import astropy.units as u

from ..analysis import (find_best_detector_position,
                        resolvingpower_per_order)
from ..math.utils import e2h
from ..source import PointSource, FixedPointing
from ..simulator import Sequence
from ..optics import (CATGrating,
                      CircleAperture, PerfectLens, RadialMirrorScatter,
                      FlatDetector)
from ..design import RowlandTorus, GratingArrayStructure


def test_detector_position():
    '''Check that the optimal detector position is found at the convergent point.'''
    n = 1000
    convergent_point = np.array([3., 5., 7.])
    pos = np.random.rand(n, 3) * 100. + 10.
    dir = - pos + convergent_point[np.newaxis, :]
    photons = Table({'pos': e2h(pos, 1), 'dir': e2h(dir, 0),
                     'energy': np.ones(n), 'polarization': np.ones(n), 'probability': np.ones(n)})
    opt = find_best_detector_position(photons)
    assert np.abs(opt.x - 3.) < 0.1

def test_resolvingpower_consistency():
    '''Compare different methods to measure the resolving power.

    This test only ensures consistency, not correctness. However, most of the
    underlying statistical functions are implemented in other packages and\
    tested there.

    This test requires a full pipeline to set up the input photons correctly
    and it thus also serves as an integration test.
    '''
    entrance = np.array([12000., 0., 0.])
    aper = CircleAperture(position=entrance, zoom=100)
    lens = PerfectLens(focallength=12000., position=entrance, zoom=100)
    rms = RadialMirrorScatter(inplanescatter=1e-4 * u.rad,
                              perpplanescatter=1e-5 * u.rad,
                              position=entrance, zoom=100)

    uptomirror = Sequence(elements=[aper, lens, rms])
    # CAT grating with blaze angle ensure that positive and negative orders
    # are defined the same way for all gratings in the GAS.
    blazeang = 1.91
    rowland = RowlandTorus(6000., 6000.)
    blazemat = transforms3d.axangles.axangle2mat(np.array([0, 0, 1]), np.deg2rad(blazeang))
    gas = GratingArrayStructure(rowland=rowland, d_element=30.,
                                x_range=[1e4, 1.4e4],
                                radius=[50, 100],
                                elem_class=CATGrating,
                                elem_args={'d': 1e-4, 'zoom': [1., 10., 10.],
                                           'orientation': blazemat,
                                           'order_selector': None},
                            )
    star = PointSource(coords=SkyCoord(23., 45., unit="degree"), flux=5. / u.s / u.cm**2)
    pointing = FixedPointing(coords=SkyCoord(23., 45., unit='deg'))
    photons = star.generate_photons(exposuretime=200 * u.s)
    p = pointing(photons)
    p = uptomirror(p)

    o = np.array([0, -3, -6])

    res1 = resolvingpower_per_order(gas, p.copy(), orders=o, detector=None)
    res2 = resolvingpower_per_order(gas, p.copy(), orders=o, detector=rowland)
    res3 = resolvingpower_per_order(gas, p.copy(), orders=o, detector=FlatDetector(zoom=1000))

    # FWHM is similar
    assert np.isclose(res1[1][0], res2[1][0], atol=0.1)
    assert np.isclose(res1[1][1], res2[1][1], atol=0.2)  # differs stronger here if fit not good
    assert np.isclose(res2[1][0], 1.8, rtol=0.1, atol=0.1)
    # Resolution of 0th order is essentially 0
    assert np.isclose(res1[0][0], 0, atol=0.5)
    assert np.isclose(res2[0][0], 0, atol=0.5)
    assert np.isclose(res3[0][0], 0, atol=0.5)
    # Resolution of higher orders is consistent and higher
    assert np.isclose(res1[0][1], res2[0][1], rtol=0.1)
    assert np.isclose(res1[0][2], res2[0][2], rtol=0.2)
    assert np.isclose(res1[0][1], res3[0][1], rtol=0.1)
    # Resolution is higher at higher orders (approximately linear for small angles)
    assert np.isclose(res1[0][2], 2 * res1[0][1], rtol=0.2)
    assert np.isclose(res2[0][2], 2 * res2[0][1], rtol=0.2)
    # No test for res3 here, since that does not follow Rowland circle.



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

class RandomArbitraryPdf(object):
    '''Take random draw from an arbitrary (and arbitrarily binned) pdf.

    The PDF is approximated as piecewice constant.
    The object is initialized with a parameters that describe the PDF, the object can
    then be called with the number of random samples that are required.

    Parameters
    ----------
    x : np.array
        **Upper** bin edge for input bins
    pdf : np.array
        Value of the pdf for each bin. ``pdf[0]`` is ignored, since the lower edge
        of that bin is undefined.
    randomize_in_bin : bool
        If ``True`` randomize the return value over each bin. If ``False`` the return
        value will be exactly the **upper** bin edge.
    sort : bool
        When calculating the cdf from a pdf that covers a large dynamical range,
        round-off errors may occur. If ``True`` this sorts the input bins in size
        (keeping an index for reverse the operation later) to avoid numerical errors.

    Examples
    --------
    First, and example with two bins (1-2 and 2-3). The probability density for each
    bin is 1 and 6, respectively (the 0.2 is ignored, because the range XXX - 1 is not
    a valid bin since the lower bound is undefined).

    >>> import numpy as np
    >>> # Set seed so this example returns exact same numbers every time
    >>> np.random.seed(0)
    >>> a = RandomArbitraryPdf(np.array([1,2,3]), np.array([0.2,1,6]))
    >>> a(10)
    array([2.79172504, 2.52889492, 2.56804456, 2.92559664, 2.07103606,
           2.0871293 , 2.0202184 , 2.83261985, 2.77815675, 2.87001215])

    As you can see, in this case the chance to draw from the 1-2 interval is only
    1/7 and it did not happen this time.

    If, instead, we want to have exact (upper) bin edges returned it looks like this:
    >>> a = RandomArbitraryPdf(np.array([1,2,3]), np.array([0,1,6]), randomize_in_bin=False)
    >>> a(10)
    array([3, 3, 3, 3, 2, 3, 3, 3, 3, 3])

    As expected, most numbers are 3, with a 2 mixed in.

    References
    ----------
    https://en.wikipedia.org/wiki/Pseudorandom_number_generator#Non-uniform_generators

    http://stackoverflow.com/questions/21100716/
    '''
    def __init__(self, x, pdf, randomize_in_bin=True, sort=True):
        if not len(x) == len(pdf):
            raise ValueError('x and pdf must have same number of elements.')
        if not np.all(np.array(pdf) >= 0):
            raise ValueError('pdf cannot have negative elements.')

        self.x = np.asarray(x)
        self.bin_width = np.hstack(([0], np.diff(x)))
        if not np.all(self.bin_width >=0):
            raise ValueError('x must be input in increasing order.')
        pdf = np.asarray(pdf) * self.bin_width
        self.sort = sort
        self.randomize_in_bin = randomize_in_bin

        # sort the pdf - otherwise bins with small numbers might be lost to round-off errors
        # idea is from http://stackoverflow.com/questions/21100716/
        if self.sort:
            self.sortindex = np.argsort(pdf)
            self.pdf = pdf[self.sortindex]
        else:
            self.pdf = pdf
        # cumulative distribution function
        self.cdf = np.cumsum(self.pdf)

    def __call__(self, N):
        """Draw from the distribution function. See docstring of class."""
        #pick numbers which are uniformly random over the cumulative distribution function
        choice = np.random.uniform(high=self.cdf[-1], size=N)
        # Now here is the difficult and comparatively expensive part:
        # We need a reverse lookup to find the bin in the cdf so that we an use it
        # to map this back to the x values of the pdf
        # np.searchsorted is a binary search with O(log n).
        index = np.searchsorted(self.cdf, choice)
        #if necessary, map the indices back to their original ordering
        if self.sort:
            index = self.sortindex[index]
        if self.randomize_in_bin:
            return self.x[index - 1] + self.bin_width[index] * np.random.rand(N)
        else:
            return self.x[index]



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy import table
import astropy.units as u

from .base import FlatOpticalElement
from ..base import GeometryError
from ..math.geometry import RectangleHole, CircularHole
from ..simulator import BaseContainer


class BaseAperture(object):
    '''Base Aperture class'''

    display = {'color': (0.0, 0.75, 0.75),
               'opacity': 0.3,
               'shape': 'triangulation'}

    @staticmethod
    def add_colpos(photons):
        '''add columns ['pos'] to photon array'''
        if 'pos' not in photons.colnames:
            photoncoords = table.Column(name='pos', length=len(photons),
                                        shape=(4,))
            photons.add_column(photoncoords)
            photons['pos'][:, 3] = 1
            photons['pos'].unit = u.mm

    @property
    def area(self):
        '''Area of the aperture.

        This does not take into account any projection effects for
        apertures that are not perpendicular to the optical axis.
        '''
        return NotImplementedError


class FlatAperture(BaseAperture, FlatOpticalElement):
    'Base class for geometrically flat apertures defined in python'

    def __call__(self, photons):
        # The last two arguments make no sense for apertures - the
        # intercoos and interpoos are assigned in specific_process_photons
        # instead of derived from intersect - but we need something here
        # to keep the interface the same as in FlatOpticalElement.
        self.add_colpos(photons)
        self.process_photons(photons, np.ones(len(photons), dtype=bool),
                             photons['pos'],
                             np.zeros((len(photons), 2)))
        return photons

    def generate_local_xy(self, n):
        '''Generate x, y in the local coordinate system

        Parameters
        ----------
        n : int
            number of x, y coordinate pairs requested

        Returns
        -------
        x, y : arrays of floats
            (x, y) coordinates of the photons in the local frame
        '''
        return NotImplementedError

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        x, y = self.generate_local_xy(intersect.sum())

        intercoos[intersect, 0] = x
        intercoos[intersect, 1] = y
        interpos[intersect, :] = self.geometry['center'] + x.reshape((-1, 1)) * self.geometry['v_y'] + y.reshape((-1, 1)) * self.geometry['v_z']
        projected_area = np.dot(photons['dir'][intersect].data,
                                - self.geometry['e_x'])
        # Photons coming "through the back" would have negative probabilities.
        # Unlikely to ever come up, but just in case we clip to 0.
        return {'probability': np.clip(projected_area, 0, 1.)}

    def outer_shape(self):
        '''Return values in Eukledian space.'''
        return self.xyz_square(self.display.get('outer_factor', 3))


class RectangleAperture(FlatAperture):
    '''Select the position where a parallel ray from an astrophysical source starts the simulation.

    '''

    default_geometry = RectangleHole

    def generate_local_xy(self, n):
        x = np.random.random(n) * 2. - 1.
        y = np.random.random(n) * 2. - 1.
        return x, y

    @property
    def area(self):
        '''Area covered by the aperture'''
        return 4 * np.linalg.norm(self.geometry['v_y']) * np.linalg.norm(self.geometry['v_z']) * u.mm**2


class CircleAperture(FlatAperture):
    '''Select the position where a parallel ray from an astrophysical source starts the simulation.

    Photons are placed in a circle. The radius of the circle is the lengths of
    its ``v_y`` vector.  At this point, the aperture must have the same zoom in
    y and z direction; it cannot be used to simulate an entrance ellipse.

    Parameters
    ----------
    phi : list of 2 floats
        Bounding angles for a segment covered by the GSA. :math:`\phi=0`
        is on the positive y axis. The segment fills the space from ``phi1`` to
        ``phi2`` in the usual mathematical way (counterclockwise).
        Angles are given in radian. Use ``phi[1] < 0`` if the segment crosses
        the y axis. (Default is the full circle.)
    r_inner : float
        Inner radius for ring-like apertures. Default is 0 (full circle). If
        `r_inner` is non-zero, plotting the aperture will fill in the inner
        region.  If this is not desired because several `CircleApertures` are
        stacked into each other, ``self.display['inner_factor']`` can be used
        to restrict the radius range where the inner disk is displayed in a
        plot.

    '''

    default_geometry = CircularHole

    def __init__(self, **kwargs):
        super(CircleAperture, self).__init__(**kwargs)
        if self.geometry['r_inner'] > np.linalg.norm(self.geometry['v_y']):
            raise ValueError('r_inner must be less than size of full aperture.')

        if not np.isclose(np.linalg.norm(self.geometry['v_y']),
                          np.linalg.norm(self.geometry['v_z'])):
            raise GeometryError('Aperture does not have the same size in y, z direction.')

    def generate_local_xy(self, n):
        phi = np.random.uniform(self.geometry.phi[0], self.geometry.phi[1], n)
        # normalize r_inner
        r_inner = self.geometry['r_inner'] / np.linalg.norm(self.geometry['v_y'])
        r = np.sqrt(np.random.uniform(r_inner**2, 1., n))

        x = r * np.cos(phi)
        y = r * np.sin(phi)
        return x, y

    @property
    def area(self):
        '''Area covered by the aperture'''
        A_circ = np.pi * (np.linalg.norm(self.geometry['v_y'])**2 - self.geometry['r_inner']**2)
        return (self.geometry.phi[1] - self.geometry.phi[0])  / (2 * np.pi) * A_circ * u.mm**2


class MultiAperture(BaseAperture, BaseContainer):
    '''Group several apertures into one class.

    Sometimes a single intrument has several physical openings where photons
    from an astrophysical source can enter, an example is XMM-Newton that
    operates three telescopes in parallel. While it is often more efficient to
    simulate these as entirely separate by running separate simulations, that
    is not always true. This class groups several apertures together.

    .. warning::

       Apertures cannot overlap. There is currently no code checking for this,
       but overlapping apertures will produce unphysical results.


    Parameters
    ----------
    elements : list
        The elements of this list are all optical elements that process photons.
    preprocess_steps : list
        The elements of this list are functions or callable objects that accept
        a photon list as input and return no output (*default*: ``[]``). All
        ``preprocess_steps`` are run before *every* aperture on just the
        photons that pass this aperture.
    postprocess_steps : list
        See ``preprocess_steps`` except that the steps are run *after* each
        aperture (*default*: ``[]``) on just the photons that passed that
        aperture.

    '''
    display = {'shape': 'container'}

    def __init__(self, **kwargs):
        self.elements = kwargs.pop('elements')
        self.id_col = kwargs.pop('id_col', 'aperture')
        self.id_num_offset = kwargs.pop('id_num_offset', 0)
        super(MultiAperture, self).__init__(**kwargs)
        for i, elem in enumerate(self.elements):
            elem.id_col = self.id_col
            elem.id_num = self.id_num_offset + i

    @property
    def area(self):
        '''Area covered by the aperture'''
        return u.Quantity([e.area for e in self.elements]).sum()

    def __call__(self, photons):
        self.add_colpos(photons)
        areas = u.Quantity([e.area for e in self.elements])
        aperid = np.digitize(np.random.rand(len(photons)), np.cumsum(areas) / self.area)
        np.random.shuffle(aperid)

        for i, elem in enumerate(self.elements):
            for p in self.preprocess_steps:
                p(photons)
            # The following line differs from "normal" optical
            # elements. In other elements "intersect" decides
            # which photons are touched, here we pass that in.
            photons = elem.process_photons(photons, aperid==i,
                                           photons['pos'],
                                           np.zeros((len(photons), 2)))
            for p in self.postprocess_steps:
                p(photons)
        return photons



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest

from ..pluecker import *
from ..utils import *

pos1d100 = np.array([1.,0,0,1])
pos1d100_a = np.array([-3., 0, 0, -3])
pos1d200 = np.array([2.,0,0,1])
pos1d333 = np.array([3. , 3, 3, 1])
pos2d000 = np.array([[0.,0,0,1],[0,0,0,2]])
pos2d100 = np.array([[1.,0,0,1],[2,0,0,2]])


# This might not actually need this many tests, but now I have them.
@pytest.mark.parametrize("in1,in2,expected", zip([pos1d100, pos1d100_a, pos2d000, pos2d100], [pos1d200, pos2d100, pos1d200, pos2d000], [1, np.zeros(2), 2*np.ones(2), np.ones(1)]))
def test_distance_point_point(in1, in2, expected):
    '''Test euklidean distance between point in homogeneous coordinates.'''
    assert np.all(distance_point_point(in1, in2) == expected)


def test_pointdir2plane():
    point = np.random.random(4) - 0.5
    # cannot be 0, and too small number might cause numeric inaccuracy
    point[3] = max(point[3], 0.1)
    direction = np.random.random(4) - 0.5
    direction[3] = 0
    p = point_dir2plane(point, direction)
    assert np.allclose(np.dot(p, point), 0)
    assert np.allclose(np.cross(p[:3], direction[:3]), 0)


def test_dir_point2line():
    '''Construct Plucker coordinates from direction and point

    This example is taken from
    Pluecker Coordinate Tutorial, by Ken Shoemake
    '''
    p = np.array([2., 3, 7])
    q = np.array([2., 1, 0])
    line = np.array([0., 2, 7, -7, 14, -4])
    assert np.all(dir_point2line(p - q, p) == line)
    # Same calculation, but 2 dim inputs
    p = np.tile(p, (2, 1))
    q = np.tile(q, (2, 1))
    lines = np.tile(line, (2, 1))
    assert np.all(dir_point2line(p - q, p) == lines)


def test_intersect_line_plane():
    q = np.array([2., 1, 0])
    line = np.array([0., 2, 7, -7, 14, -4])
    assert np.all(h2e(intersect_line_plane(line, np.array([0., 0., 1., 0.]))) == q)
    # Similar calculation but for several input lines.
    # First makes the lines
    p = np.array([[3., 5, 2], [-4, 5., 2.]])
    # Very small chance that I generate a line that's in the plane.
    # If that ever comes up in tests, hardcode something better.
    q = np.random.random((2, 3))
    lines = e_pointpoint2line(p, q)
    plane = np.array([0, -1., 0, 5])
    assert np.allclose(h2e(intersect_line_plane(lines, plane)), p)


def test_intersect_line_plane_parallel():
    '''Check results for a line that is parallel to or even in the plane.'''
    # xy plane at z = +20
    plane = np.array([0, 0, 1, -20.])
    line_in_plane = dir_point2line(np.array([1, 5, 0]), np.array([5,-500,20]))
    assert np.all(intersect_line_plane(line_in_plane, plane) == 0)

    line_parallel_plane = dir_point2line(np.array([-.1, 0., 0.]), np.ones(3))
    intersect = intersect_line_plane(line_parallel_plane, plane)
    assert intersect[3] == 0
    assert np.any(intersect != 0)



# Licensed under GPL version 3 - see LICENSE.rst
from .simulator import (SimulationSetupError,
                        BaseContainer, Sequence, Parallel,
                        ParallelCalculated,
                        KeepCol,
                        propagate,
                        Propagator
                        )



# Licensed under GPL version 3 - see LICENSE.rst
'''This module collects helper functions for visualization that are of use for several backends.

The functions here are not intended to be called directly by the user. Instead, they
refractor common tasks that are used in several visualization backends.
'''
from __future__ import division

import warnings
import numpy as np


class MARXSVisualizationWarning(Warning):
    '''Warning class for MARXS objects missing from plotting'''
    pass


def get_obj_name(obj):
    '''Return printable name for objects or functions.'''
    if hasattr(obj, 'name'):
        return obj.name
    elif hasattr(obj, 'func_name'):
        return obj.func_name
    else:
        return str(obj)


class DisplayDict(dict):
    '''A dictionary to store how an element is displayed in plotting.

    A dictionary of this type works just like a normal dictionary, except for an
    additional look-up step for keys that are not found in the dictionary itself.
    A ``DisplayDict`` is initialized with a reference to the object it describes
    and any parameters accessed from ``DisplayDict`` that is not found in the
    dictionary, will be searched in the objects geometry. This allows us to set
    any and all display settings in the ``DisplayDict`` to custamize plotting in any
    way, but for those values that are not set, fall back to the settings of the
    geometry (e.g. the shape of an object is typically taken from the geometry,
    while the color is not).

    Parameters
    ----------
    parent : `marxs.base.MarxsElement`
        Reference to the object that is described by this ``DisplayDict``
    args, kwargs: see `dict`
    '''
    def __init__(self, parent, *args, **kwargs):
        self.parent = parent
        super(DisplayDict, self).__init__(*args, **kwargs)

    def __getitem__(self, key):
        if (key not in self) and hasattr(self.parent, 'geometry'):
            try:
                return getattr(self.parent.geometry, key)
            except AttributeError:
                raise KeyError(key)
        else:
            return super(DisplayDict, self).__getitem__(key)

    def get(self, k, d=None):
        '''D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None.'''
        try:
            return self[k]
        except:
            KeyError
            return d



def plot_object_general(plot_registry, obj, display=None, **kwargs):
    '''Look up a plotting routine for an object and execute it.

    This function is not meant to be called directly by the user, instead, it
    is designed to simplify the implementation of new plotting backends.

    Parameters
    ----------
    plot_registry : dict
        Keys are the names of the shape of an object and values in this
        dictionary are functions that know how to plot this type of shape. The
        appropriate plotting function is then called with the input `obj`,
        `display` and any other keyword arguments.
        If the shape is ``"None"`` (as a string), no plotting function is
        called.
    obj : `marxs.base.MarxsElement`
        The element that should be plotted.
    display : dict of None
        Dictionary with display settings. If this is ``None``, ``obj.display``
        is used. If that is also ``None`` then the objects is skipped.
    kwargs : other keyword arguments
        These arguments are just passed through to the plotting function.

    Returns
    -------
    out : different
        The output from the plotting function that was executed is passed
        through.  Different plotting backends return different kinds of output.
    '''
    if display is None:
        if hasattr(obj, 'display') and (obj.display is not None):
            display = obj.display
        else:
            warnings.warn('Skipping {0}: No display dictionary found.'.format(get_obj_name(obj)),
                           MARXSVisualizationWarning)
            return None

    try:
        shape = display['shape']
    except KeyError:
        warnings.warn('Skipping {0}: "shape" not set in display dict.'.format(get_obj_name(obj)),
                       MARXSVisualizationWarning)
        return None

    shapes = [s.strip() for s in shape.split(';')]
    for s in shapes:
        if s == 'None':
            return None
        elif s in plot_registry:
            # turn into valid color tuple
            display['color'] = get_color(display)
            return plot_registry[s](obj, display,  **kwargs)
    else:
        warnings.warn('Skipping {0}: No function to plot {1}.'.format(get_obj_name(obj), shape),
                          MARXSVisualizationWarning)
        return None


def get_color(d):
    '''Look for color information in dictionary.

    If missing, return white.
    This function checks if the `d['color']` is a valid RGB tuple and if
    not it imports a `matplotlib.colors.ColorConverter` to convert any
    matplotlib compatible string to an RGB tuple.

    Parameters
    ----------
    d : dict
        Color information should be present in ``d['color']``.

    Returns
    -------
    color : tuple
        RGB tuple with each element in the range 0..1
    '''
    if 'color' not in d:
        return (1., 1., 1.)
    else:
        c = d['color']
        # check if this is a tuple of three floats n the range 0..1
        # or can be converted to one
        try:
            cout = tuple(c)
            if len(cout) != 3:
                raise TypeError
            for a in cout:
                if not isinstance(a, float) or (a < 0.) or (a > 1.):
                    raise TypeError
            return cout
        except TypeError:
            # It's a hex or string. Let matplotlib deal with that.
            import matplotlib.colors
            return matplotlib.colors.colorConverter.to_rgb(c)


def color_tuple_to_hex(color):
    '''Convert color tuple to hex string.

    Parameters
    ----------
    color : tuple
        tuple has three elements (rgb) that are floats betwen 0 and 1
        or ints between 0 and 255.

    Returns
    -------
    hexstring : string
        string encoding that number as hex
    '''
    if all([isinstance(a, float) for a in color]):
        if any(i < 0. for i in color) or any(i > 1. for i in color):
            raise ValueError('Float values in color tuple must be between 0 and 1.')
        out = hex(int(color[0] * 256**2 * 255 + color[1] * 256 * 255 + color[2] * 255))
    elif all([isinstance(a, int) for a in color]):
        if any(i < 0 for i in color) or any(i > 255 for i in color):
            raise ValueError('Int values in color tuple must be between 0 and 255.')
        out = hex(color[0] * 256**2 + color[1] * 256 + color[2])
    else:
        raise ValueError('Input tuple must be all float or all int.')
    # Now pad with zeros if required
    return out[:2] + out[2:].zfill(6)


def plane_with_hole(outer, inner):
    '''Triangulation of a plane with an inner hole

    This function constructs a triangulation for a plane with an inner hole, e.g.
    a rectangular plane where an inner circle is cut out.

    Parameters
    ----------
    outer, inner : np.ndarray of shape (n, 3)
        Coordinates in x,y,z of points that define the inner and outer
        boundary.  ``outer`` and ``inner`` can have a different number of
        points, but points need to be listed in the same orientation
        (e.g. clockwise) for both and the starting points need to have a
        similar angle as seen from the center (e.g. for a plane with z=0, both
        ``outer`` and ``inner`` could list a point close to the y-axis first.

    Returns
    -------
    xyz : nd.array
        stacked ``outer`` and ``inner``.
    triangles : nd.array
        List of the indices. Each row has the index of three points in ``xyz``.

    Examples
    --------
    In this example, we make a square and cut out a smaller square in the middle.

    >>> import numpy as np
    >>> from marxs.visualization.utils import plane_with_hole
    >>> outer = np.array([[-1, -1, 1, 1], [-1, 1, 1, -1], [0,0,0,0]]).T
    >>> inner = 0.5 * outer
    >>> xyz, triangles = plane_with_hole(outer, inner)
    >>> triangles
    array([[0, 4, 5],
       [0, 1, 5],
       [1, 5, 6],
       [1, 2, 6],
       [2, 6, 7],
       [2, 3, 7],
       [3, 7, 4],
       [3, 0, 4]])

    '''
    n_out = outer.shape[0]
    n_in = inner.shape[0]
    n = n_out + n_in

    triangles = np.zeros((n, 3), dtype=int)
    xyz = np.vstack([outer, inner])

    i_in = 0
    i_out = 0
    for i in range(n_out + n_in):
        if i/n >= i_in/n_in:
            triangles[i, :] = [i_out, n_out + i_in, n_out + ((i_in + 1) % n_in)]
            i_in += 1
        else:
            triangles[i, :] = [i_out, (i_out + 1) % n_out, n_out + (i_in % n_in)]
            i_out = (i_out + 1) % n_out
    return xyz, triangles


def combine_disjoint_triangulations(list_xyz, list_triangles):
    '''Combine two disjoint triangulations into one set of points

    This function combines two entirely separate triangulations into one set
    of point and triangles. Plotting the combined triangulation should have the
    same effect as plotting each triangulation separately. This function is used
    for plotting apertures where we have e.g. an open ring. This can be plotted
    as an inner circle plus an outer shape with a hole in it.

    Parameters
    ----------
    list_xyz : list of `np.array`
        Each array holds xyz values for one triangulation
    list_triangles : list of nd.array
        Each array holds the list of the indices for one triangulation.

    Returns
    -------
    xyz : nd.array
        stacked ``outer`` and ``inner``.
    triangles : nd.array
        List of the indices. Each row has the index of three points in ``xyz``.
    '''
    xyz = np.vstack(list_xyz)
    n_offset = np.cumsum([a.shape[0] for a in list_xyz])
    n_offset -= n_offset[0]
    triangles = np.vstack([list_triangles[i] + n_offset[i] for i in range(len(n_offset))])
    return xyz, triangles



# Licensed under GPL version 3 - see LICENSE.rst
from .source import (Source, PointSource,
                     RadialDistributionSource,
                     SphericalDiskSource,
                     DiskSource,
                     GaussSource,
                     SymbolFSource,
                     poisson_process,
                     )
from .pointing import FixedPointing, JitterPointing
from .labSource import FarLabPointSource, LabPointSourceCone



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from transforms3d import affines, euler

__all__ = ['generate_facet_uncertainty']

def generate_facet_uncertainty(n, xyz, angle_xyz):
    '''Generate 4d matrices that represent facet misalignment.

    Positional and rotational uncertainties are input to this function. It then
    draws randomnly from Gaussians centered on 0 (the correct position) for the displacement
    and rotation, where the :math:`\sigma` of the Gaussian is given by the numbers in the input.
    The linear displacements and angles are expressed as (4,4) matrixes suitable for use with
    homogeneous coordinates.

    Parameters
    ----------
    n : int
        Number of 4d matrixes to be calculated
    xyz : tuple of 3 floats
        accuracy of grating positioning in x, y, z (in mm) - Gaussian sigma, not FWHM!
    angle_xyz : tuple of 3 floats
        accuracy of grating positioning. Rotation around x, y, z (in rad) - Gaussian sigma, not FWHM!

    Returns
    -------
    pos_uncert : list of n (4,4) np.arrays
        Random realizations of the uncertainty
    '''
    translation = np.random.normal(size=(n, 3)) * np.asanyarray(xyz)[np.newaxis, :]
    rotation = np.random.normal(size=(n, 3)) * np.asanyarray(angle_xyz)[np.newaxis, :]
    return [affines.compose(t, euler.euler2mat(a[0], a[1], a[2], 'sxyz'), np.ones(3))
            for t, a in zip(translation, rotation)]



# Licensed under GPL version 3 - see LICENSE.rst
'''
.. todo::
   Add a mechanism for unit conversion. This expects all distances in mm and all energies in keV.
'''
import os
import numpy as np
from astropy.table import Table, Column, join
import astropy.units as u

from ..math.utils import h2e, e2h
from ..math.polarization import parallel_transport
from .base import OpticalElement, photonlocalcoords
from .aperture import BaseAperture

try:
    from _marx import lib as marx
    from _marx import ffi
    HAS_MARX = True
except ImportError:
    HAS_MARX = False


class MarxError(Exception):
    '''Error in the compiled Marx C module'''
    pass


class MarxMirror(OpticalElement, BaseAperture):
    '''Interface to MARX mirror module

    This class provides an interface to the `MARX <http://space.mit.edu/ASC/marx/>`_
    mirror module. It requires
    both the MARX source code and compiled MARX binaries as explained in the
    :ref:`Installation instructions for MARXS <sect-installmarxccode>`
    When this model
    is initialized, it requires the path and filename to a MARX setup file.

    The default geometry is such that the focal point is at the origin of the
    coordinate system and the optical axis is along the x-axis, such that
    photons travel from +infinity towards the origin.

    A MarxMirror object does not only act as a mirror, it fulfills the function
    of an aperture at the same time, no further aperture class should be used
    in the same simulation.

    The model reads the relevant MARX mirror model parameters from that file;
    in this way it supports all mirror models that the traditional MARX version
    implements, most notable the "HRMA" and "IXO" mirrors. Details on the
    "HRMA" mirror (which is used to simulate Chandra) can be found at
    http://space.mit.edu/ASC/marx/indetail/hardwaremodel.html#hrma-model . The
    "IXO" mirror model is a simplified version with some generalizations that
    works very similar in general, but is not documented because (from a MARX
    standpoint) it is only meant for developers.

    Parameters
    ----------
    parfile : string
        Path and filename of a MARX parameter file that sets all MARX
        parameters for the mirror model.
    '''

    def __init__(self, parfile, **kwargs):
        # If the state shared between different object that use the s
        # same C module? In that case I need to add a lock so that only
        # one object of this class can exist at any one time.
        if not HAS_MARX:
            raise MarxError('MARX C code is not available. Please see installation instructions.')
        if not os.path.isfile(parfile):
            raise IOError('MARX parameter file {0} does NOT exist.'.format(parfile))
        else:
            self.parfile = parfile
        self.cparfile = marx.pf_open_parameter_file(bytes(parfile, 'utf8'), bytes('r', 'utf8'))
        out = marx.marx_mirror_init(self.cparfile)
        if out < 0:
            raise MarxError('Mirror cannot be initialized. Probably missing parameters or syntax error in {0}.'.format(parfile))

        super(MarxMirror, self).__init__(**kwargs)

    @staticmethod
    def _table2c(photons):
        # Arrays assigned here in python need to keep a reference
        # somewhere, otherwise they would be garbage collected
        # and the pointer would suddenly be invalid.
        # To do so, this function adds them to the photons table header.
        # If this turns out to be too slow, then I have to move it to
        # Cython that can access the native numpy array memory.
        n = len(photons)
        cp = ffi.new('Marx_Photon_Attr_Type[]', n)
        keep_cffi_pointers = {}
        keep_cffi_pointers['cp'] = cp
        for i in range(n):
            cp[i].energy = photons['energy'][i]
            pos = h2e(photons['pos'][i])
            cp[i].x.x = pos[0]
            cp[i].x.y = pos[1]
            cp[i].x.z = pos[2]
            dir = h2e(photons['dir'][i])
            cp[i].p.x = dir[0]
            cp[i].p.y = dir[1]
            cp[i].p.z = dir[2]
            cp[i].arrival_time = photons['time'][i]
        if 'tag' not in photons.colnames:
            photons.add_column(Column(name='tag', data=np.arange(n)))
        for i, t in zip(range(n), photons['tag']):
            cp[i].tag = t

        c_photon_list = ffi.new('Marx_Photon_Type *')
        c_photon_list.attributes = cp
        c_photon_list.n_photons = n
        c_photon_list.max_n_photons = n
        c_photon_list.total_time = np.max(photons['time'])
        c_photon_list.start_time = np.min(photons['time'])

        sorted_index = np.argsort(photons['energy'].data)
        sorted_index = np.ascontiguousarray(sorted_index, dtype=np.uintc)
        keep_cffi_pointers['sorted_index'] = sorted_index  # keep alive
        c_photon_list.sorted_index = ffi.cast('unsigned int*', sorted_index.ctypes.data)

        sorted_energies = np.sort(photons['energy'].data)
        sorted_energies = np.ascontiguousarray(sorted_energies, dtype=np.float)
        keep_cffi_pointers['sorted_energies'] = sorted_energies  # keep alive
        c_photon_list.sorted_energies = ffi.cast('double*', sorted_energies.ctypes.data)

        c_photon_list.num_sorted = n  # all photons in list are valid
        # Fields not mentioned here are irrelevant for the mirror
        # module and can have any value they are initialized to.

        return c_photon_list, keep_cffi_pointers

    @staticmethod
    def _c2table(c_photon_list):
        '''

        We'll first collect all the data in numpy arrays and make
        a new table. Might not be the most efficient solution,
        but should work for now.
        Make work first, improve later.

        To-Do: keep absorbed?
        '''
        n_valid = c_photon_list.num_sorted
        cp = c_photon_list.attributes

        energy = np.empty(n_valid)
        time = np.empty(n_valid)
        tag = np.empty(n_valid, dtype=int)
        pos = np.empty((n_valid, 3))
        dir = np.empty((n_valid, 3))
        unreflected = np.empty(n_valid, dtype=bool)
        vblocked = np.empty(n_valid, dtype=bool)
        shell = np.empty(n_valid, dtype=int)

        for i in range(n_valid):
            energy[i] = cp[i].energy
            pos[i, 0] = cp[i].x.x
            pos[i, 1] = cp[i].x.y
            pos[i, 2] = cp[i].x.z
            dir[i, 0] = cp[i].p.x
            dir[i, 1] = cp[i].p.y
            dir[i, 2] = cp[i].p.z
            time[i] = cp[i].arrival_time
            tag[i] = cp[i].tag
            unreflected[i] = cp[i].flags & marx.PHOTON_UNREFLECTED
            vblocked[i] = cp[i].flags & marx.PHOTON_MIRROR_VBLOCKED
            shell[i] = cp[i].mirror_shell

        pos = e2h(pos, 1)
        dir = e2h(dir, 0)
        photons = Table([pos, dir, energy, time, tag,
                         unreflected, vblocked, shell],
                        names=['pos', 'dir', 'energy', 'time', 'tag',
                               'unreflected', 'mirror_vblocked', 'mirror_shell'])
        return photons

    @photonlocalcoords
    def _process_photons_in_c(self, photons, verbose):
        '''Wrap the MARX C module'''
        c_photon_list, keep_cffi_pointers = self._table2c(photons)
        out = marx.marx_mirror_reflect(c_photon_list, verbose)
        if out != 0:
            raise MarxError('Error in marx_mirror_reflect.')
        return self._c2table(c_photon_list)

    def __call__(self, photons_in, verbose=0):
        self.add_colpos(photons_in)
        new_photons = self._process_photons_in_c(photons_in, verbose)
        photons = join(new_photons, photons_in, keys='tag',
                       uniq_col_name='{col_name}{table_name}',
                       table_names=['', '_beforemirror'])
        # Probability column needs special treatment, because the probability
        # is not just replaced with the new number, it's multiplicative.
        # First, match tag in to tag after join (which may have fewer photons)
        insorted = np.argsort(photons_in['tag'])
        ypos = np.searchsorted(photons_in['tag'][insorted], photons['tag'])
        indices = insorted[ypos]
        photons['probability'] *= photons_in['probability'][indices]
        photons['probability'][photons['unreflected'] | photons['mirror_vblocked']] = 0
        pol = parallel_transport(photons_in['dir'][insorted], photons['dir'],
                                 photons_in['polarization'][insorted])
        photons['polarization'] = pol
        return photons

    @property
    def area(self):
        '''Area of the aperture.

        This does not take into account any projection effects for
        apertures that are not perpendicular to the optical axis.
        '''
        return marx.Marx_Mirror_Geometric_Area * u.cm**2





# Turn into a test?
# pf = marx.pf_open_parameter_file('./marxs/optics/hrma.par', 'r')
# out = marx.marx_mirror_init(pf)
# out should be 0, but I get 2 ?!?
# Will continue and leave that mystery as a mystery for now, but I bet I'll
# need to get back to it.



"""
This is an Astropy affiliated package.
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

# For egg_info test builds to pass, put package imports here.
if not _ASTROPY_SETUP_:
    pass


'''Convert from photon energy in keV to wavelength in mm'''
energy2wave = 1.2398419292004202e-06



# Licensed under GPL version 3 - see LICENSE.rst
import os
import numpy as np

from astropy.table import Table
from astropy.coordinates import SkyCoord
import astropy.units as u

from ... import chandra
from ....source import PointSource, FixedPointing
#from .....optics import MarxMirror


CWD = os.getcwd()
TEST_DIR = os.path.dirname(__file__)


def setup_function(function):
    os.chdir(TEST_DIR)

def teardown_function(function):
    os.chdir(CWD)

def test_ditherpattern():
    '''check that the dither pattern generated with consistent with MARX.'''
    mypointing = chandra.LissajousDither(coords=SkyCoord(212.5, -33., unit='deg'),
                                         roll=15. * u.deg)
    time = np.arange(1000)
    coords = np.rad2deg(mypointing.pointing(time))
    masol = Table.read('sim_asol.fits')
    assert np.allclose(coords[:, 0], np.interp(time, masol['time']-masol['time'][0], masol['ra']))
    assert np.allclose(coords[:, 1], np.interp(time, masol['time']-masol['time'][0], masol['dec']))
    assert np.allclose(coords[:, 2], np.interp(time, masol['time']-masol['time'][0], masol['roll']))

def test_stationary_pointing():
    '''Constant pointing can also be realized through a Lissajous with amplitude=0.'''
    mysource = PointSource(coords=SkyCoord(30., 30., unit="deg"))
    fixedpointing = FixedPointing(coords=SkyCoord(30., 30., unit=u.degree), roll=15. * u.degree)
    lisspointing = chandra.LissajousDither(coords=SkyCoord(30., 30., unit=u.degree),
                                           roll=15. * u.degree, DitherAmp=np.zeros(3)*u.degree)

    photons = mysource.generate_photons(2 * u.s)
    fixedphotons = fixedpointing(photons.copy())
    lissphotons = lisspointing(photons.copy())

    assert np.allclose(fixedphotons['dir'], lissphotons['dir'])

def test_detector_coordsystems():
    '''Compare detector coordinates for known photon direction with CIAO dmcoords.

    Currently, I don't aim to make this very precise - within a pixel or so is fine.
    To make this more precise, there are several places to go:
    The most obvious one is the size of the pixel - currently the number of pixels times
    the pixel size does now match the length of the chip precisely.
    '''
    mysource = PointSource(coords=SkyCoord(30., 30., unit="deg"))
    mypointing = chandra.LissajousDither(coords=SkyCoord(30., 30., unit='deg'), roll=15. * u.degree)
    # marxm = MarxMirror('./marxs/optics/hrma.par', position=np.array([0., 0,0]))
    acis = chandra.ACIS(chips=[0,1,2,3], aimpoint=chandra.AIMPOINTS['ACIS-I'])

    photons = mysource.generate_photons(5 * u.s)
    photons = mypointing(photons)
    # We want reproducible direction, so don't use mirror, but set direction by hand
    # photons = marxm(photons)
    # photons = photons[photons['probability'] > 0]
    photons['pos'] = np.array([[100., 0, 0, 1],[100, 10, 0, 1], [100, 0, 10, 1],
                               [100, -10, 0, 1], [100, 0, -10, 1]])

    photons.meta['MISSION'] = ('AXAF', 'Mission')
    photons.meta['TELESCOP'] = ('CHANDRA', 'Telescope')

    photons = acis(photons)
    # We need the asol to run CIAO.
    # I've done that already and the expected numbers are hard-coded in the assert statement.
    # Here are commands I used:
    # mypointing.write_asol(photons, 'asol.fits')
    # chandra.Chandra.write_evt('photons.fits')

    atol = 1.
    # This is the starting point
    assert np.allclose(photons['detx'], [4096.5, 4513.2, 4096.5, 3679.8, 4096.5], atol=atol)
    assert np.allclose(photons['dety'], [4096.5, 4096.5, 3679.8, 4096.5, 4513.2], atol=atol)
    # I took these numbers and types them into dmcoords.
    # Thus, if these numbers change when increasing the accuracy of the pixel size or the
    # position of the chip edges, ... then all the numbers below need to be changed, too.
    assert np.all(photons['CCD_ID'] == [3, 3, 1, 2, 3])
    assert np.allclose(photons['chipx'], [984.34, 984.34, 355.7, 40.91, 566.56], atol=atol)
    assert np.allclose(photons['chipy'], [994.8, 577.02, 994.23, 658.95, 993.75], atol=atol)
    assert np.allclose(photons['x'], [4096.5, 4499., 3988.5, 3693.4, 4204.34], atol=atol)
    assert np.allclose(photons['y'], [4096.5, 3988.65, 3693.4, 4204.5, 4499.], atol=atol)
    assert np.allclose(photons['tdetx'], [4137.2, 4555, 4137.77, 3720, 4138.25], atol=atol)


def test_ACIS_pixel_number():
    '''The pixel size and chip size given in the docs are not consistent.

    Make sure we end up with 1024*1024 chips.
    '''
    acis = chandra.ACIS(chips=[0, 1, 2, 3], aimpoint=chandra.AIMPOINTS['ACIS-I'])
    for i in range(3):
        assert acis.elements[i].npix == [1024, 1024]

    acis = chandra.ACIS(chips=[4, 5, 6, 7, 8, 9], aimpoint=chandra.AIMPOINTS['ACIS-I'])
    for i in range(5):
        assert acis.elements[i].npix == [1024, 1024]



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table
import pytest

from ..simulator import Sequence, SimulationSetupError, Parallel, KeepCol
from ..optics import ThinLens, FlatGrating, OrderSelector

def test_pre_post_process():
    '''test pre-processing and post-processing in sequences'''

    def set_energy2(photons):
        photons['energy'] = 2
        return photons

    def set_energy3(photons):
        photons['energy'] = 3
        return photons

    def process(photons):
        photons.meta['mean_en'].append(photons['energy'].mean())

    photons = Table({'energy': [1,1,1]})
    photons.meta['mean_en'] = []

    seq_pre = Sequence(elements=[set_energy2, set_energy3], preprocess_steps=[process])
    tpre = seq_pre(photons.copy())
    assert np.all(tpre['energy'] == 3)
    assert np.all(tpre.meta['mean_en'] == [1, 2])

    seq_post = Sequence(elements=[set_energy2, set_energy3], postprocess_steps=[process])
    tpost = seq_post(photons.copy())
    assert np.all(tpost['energy'] == 3)
    assert np.all(tpost.meta['mean_en'] == [2, 3])

def test_badinput():
    '''Inputs must be callable.'''
    with pytest.raises(SimulationSetupError) as e:
        seq = Sequence(elements=[5])
    assert "is not callable" in str(e.value)

    with pytest.raises(SimulationSetupError) as e:
        seq = Sequence(elements=[ThinLens], preprocess_steps=[5])
    assert "is not callable" in str(e.value)

# Tests for Parallel could be here, but Parallel is a fairly general container.
# It is easier to test a specific case - and the tests for GratingArrayStructure do that
# for many functions or Parallel.

def test_list_elemargs():
    '''Input to elemargs can be single or list.'''
    p0 = Parallel(elem_class=FlatGrating,
                  elem_pos={'position': [np.zeros(3), np.ones(3)]},
                  elem_args={'order_selector': OrderSelector(np.arange(-3, 4)), 'd': [0.001, 0.002], 'zoom': [3,3]},
                 )
    # equivalent way of inputting things
    p1 = Parallel(elem_class=FlatGrating,
                  elem_pos={'position': [np.zeros(3), np.ones(3)], 'zoom': [3, 3]},
                  elem_args={'order_selector': OrderSelector(np.arange(-3, 4)), 'd': [0.001, 0.002]},
                 )
    # yet another way
    p2 = Parallel(elem_class=FlatGrating,
                  elem_pos={'position': [np.zeros(3), np.ones(3)]},
                  elem_args={'order_selector': OrderSelector(np.arange(-3, 4)), 'd': [0.001, 0.002], 'zoom': 3},
                  )
    for p in [p0, p1, p2]:
        assert len(p.elements) == 2
        for e in p.elements:
            assert np.linalg.norm(e.geometry['v_y']) == 3
        assert p.elements[0]._d == 0.001
        assert p.elements[1]._d == 0.002

    # but this does not work: in elem_pos all entries have to be lists
    with pytest.raises(ValueError) as e:
        p4 = Parallel(elem_class=FlatGrating,
                      elem_pos={'position': [np.zeros(3), np.ones(3)], 'zoom': 3},
                      elem_args={'order_selector': OrderSelector(np.arange(-3, 4)), 'd': [0.001, 0.002]},
                      )
    assert 'All elements in elem_pos must have the same number' in str(e.value)


def test_keepcols():
    '''Check keep cols fir very simple case.'''
    t = Table({'a': [1,2]})
    def double_a(t):
        t['a'] *= 2
        return t

    keeper = KeepCol('a')
    s = Sequence(elements=[double_a, double_a], preprocess_steps=[keeper])
    t = s(t)
    assert np.all(np.hstack(keeper.data) == [1, 2, 2, 4])



# Licensed under GPL version 3 - see LICENSE.rst
'''`Mayavi <http://docs.enthought.com/mayavi/mayavi/>`__ plotting backend

Mayavi is a python package for interactive 3-D displays that uses VTK underneath.
Fuctions in this module display rays or objects in 3D using mayavi. Each of them
requires a ``mayavi.core.scene.Scene`` instance as input and returns a mayavi object
(or a list of those), e.g. a ``mayavi.visual.Box`` instance.

The plotting routines attempt to find all valid OpenGL properties by name in the
``display`` dictionaries and apply those to the plotted object.
'''
from __future__ import absolute_import

import numpy as np
from astropy.utils.decorators import format_doc

from ..math import utils as mutils
from . import utils

# The following import fails on headless servers (Travis, readthedocs).
# from mayavi import mlab
# Thus, I've moved the import statement into the individual functions, so that
# this module can still be imported when Travis or readthedocs build the documentation.

doc_plot='''
    {__doc__}

    Parameters
    ----------
    obj : `marxs.base.MarxsElement`
        The element that should be plotted.
    display : dict of None
        Dictionary with display settings.
    viewer : ``mayavi.core.scene.Scene instance``.
        If None, the source is not added
        to any figure, and will be added automatically by the modules or filters.
        If False, no figure will be created by modules or filters applied to the
        source: the source can only be used for testing, or numerical algorithms,
        not visualization.

    Returns
    -------
    out : mayavi object
        Return the result of a mayavi plotting method.
'''

@format_doc(doc_plot)
def container(obj, display=None, viewer=None):
    '''Recursively plot objects containted in a container.'''
    return [plot_object(e, display=None, viewer=viewer) for e in obj.elements]

@format_doc(doc_plot)
def triangulation(obj, display, viewer=None):
    '''Plot a plane with an inner hole such as an aperture.'''
    from mayavi import mlab

    xyz, triangles = obj.geometry.triangulate(display)
    t = mlab.triangular_mesh(xyz[:, 0], xyz[:, 1], xyz[:, 2], triangles, color=display['color'])
    return t

@format_doc(doc_plot)
def surface(surface, display, viewer=None):
    '''Plot a parametric surface.

    The parameter boundaries are taken from the ``coo1`` and ``coo2`` in the
    display dictionary. The plotting routine is generic. It calls the
    ``parametric_surface()`` method of the object that is plotted; see there
    for a detailed description of parameters.
    '''
    from mayavi import mlab

    xyz = surface.geometry.parametric_surface(display.get('coo1', None),
                                              display.get('coo2', None),
                                              display)
    xyz = mutils.h2e(xyz)
    x = xyz[..., 0]
    y = xyz[..., 1]
    z = xyz[..., 2]
    m = mlab.mesh(x, y, z, figure=viewer, color=display['color'])
    return m

@format_doc(doc_plot)
def box(obj, display, viewer=None):
    '''Plot a rectangular box for an object.

    By default, the box extends in x,y, and z direction. The display keyword
    "box-half" can be used to show only one half of the box, e.g. "+x" would
    show  the full extend in y and z direction, but only the lower half in
    x direction (such that rays coming from the +x direction are
    visible up to the interaction point).
    Use this for elements such as mirrors or detectors where
    photon interaction happens on the surface, not in the substrate.
    '''
    from mayavi import mlab

    corners = np.array([[-1, -1, -1], [-1,+1, -1],
                        [-1, -1,  1], [-1, 1,  1],
                        [ 1, -1, -1], [ 1, 1, -1],
                        [ 1, -1, +1], [ 1, 1, +1]])
    if 'box-half' in display:
        # write in a way that it works with any value for that keyword
        try:
            if display['box-half'][0] == '+':
                factor = +1
            elif display['box-half'][0] == '-':
                factor = -1
            else:
                factor = 0
            xyz = {'x': 0, 'y': 1, 'z': 2}
            if display['box-half'][1] in xyz:
                j = xyz[display['box-half'][1]]
                corners[corners[:, j] == factor, j] = 0
        except:
            pass
    triangles = [(0,2,6), (0,4,6), (0,1,5), (0,4,5), (0,1,3), (0,2,3),
                 (7,3,2), (7,6,2), (7,3,1), (7,5,1), (7,6,4), (7,5,4)]
    corners = np.einsum('ij,...j->...i', obj.pos4d, mutils.e2h(corners, 1))
    b = mlab.triangular_mesh(corners[:,0], corners[:,1], corners[:,2],
                             triangles, color=display['color'])
    return b

@format_doc(doc_plot)
def cylinder(obj, display, viewer=None):
    '''Plot a cylinder for an object.

    The radius is taken from the y dimensions, the length of the tube from
    the x dimension.
    The display keyword "tube-sides" (default: 20) sets the number of vortices
    that are used to approximate the curve.
    '''
    from mayavi import mlab

    x0 = obj.geometry['center'] - obj.geometry['v_x']
    x1 = obj.geometry['center'] + obj.geometry['v_x']
    c = mlab.plot3d([x0[0], x1[0]], [x0[1], x1[1]], [x0[2], x1[2]],
                    color=display['color'],
                    tube_radius=np.linalg.norm(obj.geometry['v_y']),
                    tube_sides=display.get('tube_sides', 20))
    return c

def plot_rays(data, scalar=None, viewer=None,
              kwargssurface={'colormap': 'Accent', 'line_width': 1, 'opacity': .4}):
    '''Plot lines for simulated rays.

    Parameters
    ----------
    data : np.array of shape(n, N, 3)
        where n is the number of rays, N the number of positions per ray and
        the last dimension is the (x,y,z) of an Eukledian position vector.
    scalar : None or nd.array of shape (n,) or (n, N)
        This quantity is used to color the rays. If ``None`` all rays will have
        the same color. If it has n elements, each ray will have exactly one
        color (e.g. color according to the energy of the ray), if it has n*N
        elements, rays will be multicolored.
    viewer : ``mayavi.core.scene.Scene instance``.
       If None, the source is not added to any figure, and will be added
        automatically by the modules or filters.  If False, no figure will be
        created by modules or filters applied to the source: the source can
        only be used for testing, or numerical algorithms, not visualization.
    kwargssurface : dict
        keyword arguments for ``mayavi.mlab.pipeline.surface``

    Returns
    -------
    out : mayavi ojects
        This just passes through the information returned from the mayavi calls.

    '''
    from mayavi import mlab

    # The number of points per line
    N = data.shape[1]
    # number of lines
    n = data.shape[0]

    if scalar is None:
        s = np.zeros(n * N)
    elif scalar.shape == (n, ):
        s = np.repeat(scalar, N)
    elif scalar.shape == (n, N):
        s = scalar.flatten()
    else:
        raise ValueError('Scalar quantity for each point must have shape ({0},) or ({0}, {1})'.format(n, N))

    x = data[:,:, 0].flatten()
    y = data[:,:, 1].flatten()
    z = data[:,:, 2].flatten()

    a = np.arange(n * N).reshape((-1, N))[:, :-1].flatten()
    b = a + 1
    connections = np.vstack([a,b]).T

    # Create the points
    src = mlab.pipeline.scalar_scatter(x, y, z, s, figure=viewer)

    # Connect them
    src.mlab_source.dataset.lines = connections
    src.update()

    # The stripper filter cleans up connected lines
    lines = mlab.pipeline.stripper(src, figure=viewer)

    # Finally, display the set of lines
    surface = mlab.pipeline.surface(lines, figure=viewer, **kwargssurface)

    return src, lines, surface

plot_registry = {'triangulation': triangulation,
                 'surface': surface,
                 'box': box,
                 'container': container,
                 'cylinder': cylinder,
                 }


def plot_object(obj, display=None, viewer=None, **kwargs):
    '''Plot any marxs object with using Mayavi as a backend.

    This method will inspect the object that is passed in and select the
    correct plotting method for its shape. The object is added to the
    mayavi scene specified in the ``viewer``.


    Parameters
    ----------
    obj : `marxs.base.MarxsElement`
        The element that should be plotted.
    display : dict of None
        Dictionary with display settings. If this is ``None``, ``obj.display``
        is used. If that is also ``None`` then the objects is skipped.
    viewer : ``mayavi.core.scene.Scene``
        If None, the source is not added to any figure, and will be added
        automatically by the modules or filters.  If False, no figure will be
        created by modules or filters applied to the source: the source can
        only be used for testing, or numerical algorithms, not visualization.
    kwargs
        All other parameters will be passed on to the individual plotting
        method.

    Returns
    -------
    out : mayavi object
        Return the result of a mayavi plotting method.

    '''
    kwargs['viewer'] = viewer
    out = utils.plot_object_general(plot_registry, obj, display, **kwargs)

    if (out is not None) and not isinstance(out, list):
        # containers return list, but properties are already set there
        display = display or obj.display
        # No safety net here like for color converting to a tuple.
        # If the advanced properties are set you are on your own.
        prop = out.module_manager.children[0].actor.property
        for n in prop.trait_names():
            if n in display:
                setattr(prop, n, display[n])
    return out



# Licensed under GPL version 3 - see LICENSE.rst
'''This module contains filters, e.g. an optical blocking filter or CCD contamination.
'''
import numpy as np

from .base import OpticalElement, FlatOpticalElement
from ..base import MarxsElement
from ..math.geometry import NoGeometry

class GlobalEnergyFilter(MarxsElement):
    '''Energy dependent filter that globally affects all photons.

    This element is used on all photons in the list, there is no geometrical
    position associated with it. Consequently, there is no update of the position
    or direction for each photon.
    Use this element for global filters, that are not directly associated with any particular
    physical object, e.g. to apply a energy based mirror efficiency after
    passing the photons through one of the perfect efficiency mirror models.

    Parameters
    ----------
    filterfunc : callable
        A function that calculates the probability for each photon to pass
        through the filter based on the photon energy in keV. The function
        signature should be ``p = func(en)``, where ``p, en`` are 1-d arrays
        of floats with the same number of elements.

    Examples
    --------
    >>> from scipy.interpolate import interp1d
    >>> from marxs.optics import GlobalEnergyFilter
    >>> energygrid = [.1, .5, 1., 2., 5.]
    >>> filtercurve = [.1, .5, .9, .9, .5]
    >>> f = interp1d(energygrid, filtercurve)
    >>> blockingfilter = GlobalEnergyFilter(filterfunc=f)

    See Also
    --------
    marxs.optics.filter.EnergyFilter
    '''

    def __init__(self, **kwargs):
        self.filterfunc = kwargs.pop('filterfunc')
        super(GlobalEnergyFilter, self).__init__(**kwargs)

    def __call__(self, photons):
        p =  self.filterfunc(photons['energy'])
        if np.any(p < 0.) or np.any(p > 1.):
            raise ValueError('Probabilities returned by filterfunc must be in interval [0, 1].')
        photons['probability'] *= p
        return photons


class EnergyFilter(FlatOpticalElement):
    '''Energy dependent filter with position, size etc.

    Parameters
    ----------
    filterfunc : callable
        A function that calculates the probability for each photon to pass
        through the filter based on the photon energy in keV. The function
        signature should be ``p = func(en)``, where ``p, en`` are 1-d arrays
        of floats with the same number of elements.

    Examples
    --------
    >>> from scipy.interpolate import interp1d
    >>> from marxs.optics import EnergyFilter
    >>> energygrid = [.1, .5, 1., 2., 5.]
    >>> filtercurve = [.1, .5, .9, .9, .5]
    >>> f = interp1d(energygrid, filtercurve)
    >>> blockingfilter = EnergyFilter(filterfunc=f, position=[4, 1, 0], zoom=4)

    See Also
    --------
    marxs.optics.filter.GlobalEnergyFilter
    '''

    display = {'color': (1.0, 0., 0.),
               'opacity': 0.5,
               'shape': 'box',
    }
    def __init__(self, **kwargs):
        self.filterfunc = kwargs.pop('filterfunc')
        super(EnergyFilter, self).__init__(**kwargs)

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        p =  self.filterfunc(photons['energy'][intersect])
        if np.any(p < 0.) or np.any(p > 1.):
            raise ValueError('Probabilities returned by filterfunc must be in interval [0, 1].')
        return {'probability': p}



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename
from transforms3d.affines import compose

from ...optics import FlatGrating, OrderSelector
from ...simulator import Parallel

class HETG(Parallel):

    id_col = 'facet'

    def __init__(self, **kwargs):
        self.hess = Table.read(get_pkg_data_filename('HESSdesign.rdb'))
        '''The HESS datafile is commented very well inside the rdb file.
        Here, I just need to make a note about the relation of the coordinate systems:
        The vectors that define the facet edges are called x and y in the rdb file.
        In the MARXS global coordinate system these are y and z respectively, so
        uxf -> y and uyg -> z.

        Datafile from: http://space.mit.edu/HETG/hess/basic.html
        '''
        kwargs['elem_pos'] = self.calculate_elempos()
        kwargs['elem_class'] = FlatGrating
        # Gratings are defined in order MEG, then HEG
        d = [4001.95 * 1e-7] * 192 + [2000.81 * 1e-7] * 144
        ul = np.vstack([self.hess[s+'ul'].data for s in 'xyz'])
        uyf = np.vstack([self.hess[s + 'uyf'].data for s in 'xyz'])
        uxf = np.vstack([self.hess[s + 'uxf'].data for s in 'xyz'])
        groove_ang = -np.arctan2(np.einsum('i...,i...->...', ul, uxf),
                                 np.einsum('i...,i...->...', ul, uyf))
        kwargs['elem_args'] = {'order_selector': OrderSelector(np.arange(-3, 4)),
                               'd': d, 'name': list(self.hess['hessloc']),
                               'groove_angle': groove_ang.tolist()}
        super(HETG, self).__init__(**kwargs)

        # I would like to put the center of the HETG on its hinge point.
        # Unfortunately, I don't know that hinge point very well, so I'm guessing a little.
        move = np.eye(4)
        move[:, 3] = [8600., 580 * np.cos(np.deg2rad(30)),
                      580 * np.sin(np.deg2rad(30)), 1.]

    def calculate_elempos(self):
        '''Read position of facets from file.

        Based on the positions, other grating parameters are chosen that differ between
        HEG and MEG, e.g. grating constant.
        '''
        # take x,y,z components of various vectors and put in numpy arrays
        # Note: This makes (3, N) arrays as opposed to (N, 3) arrays used in other parts of MARXS.
        # Note: Sometimes an angle has a minus. Beware of active / passive rotations.
        hess = self.hess

        cen = np.vstack([hess[s+'c'].data for s in 'xyz'])
        cen[0, :] += (339.909 - 1.7) * 25.4

        # All those vectors are already normalized
        facnorm = np.vstack([hess[s + 'u'].data for s in 'xyz'])
        uxf = np.vstack([hess[s + 'uxf'].data for s in 'xyz'])
        uyf = np.vstack([hess[s + 'uyf'].data for s in 'xyz'])

        pos4ds = []
        for i in range(facnorm.shape[1]):
            if int(self.hess['hessloc'][i][0]) <= 3:
                # Size of MEG facets
                zoom = np.array([1, 1.035*25.4/2, 1.035*25.4/2])
            else:
                # Size of HEG facets
                zoom = np.array([1, 0.920*25.4/2, 0.920*25.4/2])

            rot = np.eye(3)
            rot[:, 0] = facnorm[:, i]
            rot[:, 1] = uxf[:, i]
            rot[:, 2] = uyf[:, i]
            pos4d = compose(cen[:, i], rot, zoom)
            pos4ds.append(pos4d)
        return pos4ds



import numpy as np
from astropy.table import Table
import pytest

from ..simulator import KeepCol, Sequence, BaseContainer
from ...optics import FlatDetector

f1 = FlatDetector()
f2 = FlatDetector()
f3 = FlatDetector()
s_l2 = Sequence(elements=[f2, f3])

mission = Sequence(elements=[f1, s_l2])

def test_seach_all():
    '''Check full search finding all elements.'''
    assert [mission, s_l2] == mission.elements_of_class(Sequence)
    assert [] == mission.elements_of_class(BaseContainer)
    assert [mission, s_l2] == mission.elements_of_class(BaseContainer,
                                                        subclass_ok=True)
    assert [f1, f2, f3] == mission.elements_of_class(FlatDetector)
    assert [f1, f2, f3] == mission.elements_of_class(FlatDetector, subclass_ok=True)

def test_seach_top():
    '''Check top search finding all elements.'''
    assert [mission] == mission.elements_of_class(Sequence, stop_at_first=True)
    assert [] == mission.elements_of_class(BaseContainer, stop_at_first=True)
    assert [mission] == mission.elements_of_class(BaseContainer,
                                                  subclass_ok=True,
                                                  stop_at_first=True)
    assert [f1, f2, f3] == mission.elements_of_class(FlatDetector,
                                                     stop_at_first=True)
    assert [f1, f2, f3] == mission.elements_of_class(FlatDetector,
                                                     subclass_ok=True,
                                                     stop_at_first=True)


def test_format_saved_positions():
    '''Reformat saved positions and drop nearly identical values.'''
    pos0 = np.arange(20).reshape(5,4)
    pos1 = pos0 + 1
    pos2 = pos1 + 1e-4
    pos = KeepCol('testcase')
    pos.data = [pos0, pos1, pos2]

    d = pos.format_positions(atol=1e-2)
    assert d.shape == (5, 2, 3)
    assert np.allclose(d[0, 0, :], [0, 1./3, 2./3])

    '''Not drop values'''
    d = pos.format_positions()
    assert d.shape == (5, 3, 3)

def test_empty_format_saved_positions():
    '''If the input contains no data, an error should be raised.'''
    a = KeepCol('testcase')
    with pytest.raises(ValueError) as e:
        d = a.format_positions()

    assert 'contains no data' in str(e.value)

def test_format_save_positions_fails():
    '''Can only be used when data is in homogenous coordiantes.'''
    a = KeepCol('testcase')
    a.data = [np.arange(5), np.arange(5)]
    with pytest.raises(ValueError) as e:
        d = a.format_positions()

    assert 'homogeneous coordinates' in str(e.value)

def test_to_array():
    a = KeepCol('testcase')
    col1 = Table([[1]], names=['testcase'])
    col2 = Table([[4]], names=['testcase'])
    a(col1)
    a(col2)
    assert np.all(np.sqrt(a) == [[1.], [2.]])



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np


def e2h(e, w):
    '''Convert Euclidean coordinates to homogeneous coordinates

    Parameters
    ----------
    e : np.array
        Input Euclidean coordinates. This can be multidimensional, but the last
        dimension must be of size 3.
    w : float
        ``0`` for directions (points at infinity)
        ``1`` for positions (points in Euclidean space)

    Returns
    -------
    h : np.array
        Homogeneous coordinates. Same shape as ``e`` except that the last
        dimension is now has 4 elements.
    '''
    if not ((w == 0) or (w == 1)):
        raise ValueError('w must be 0 or 1.')
    shape = list(e.shape)
    shape[-1] += 1
    h = np.empty(shape)
    h[..., :3] = e
    h[..., 3] = w
    return h


def h2e(h):
    '''Convert homogeneous coordinates to Euclidean coordinates

    This functions works for points at infinity and for points in real
    euclidean space, but it expects that each time it is called ``h`` contains
    only the one or the other but not a mixture of both.

    Parameters
    ----------
    h : np.array
        Input homogeneous coordinates. This can be multidimensional, but the
        last dimension must be of size 4.

    Returns
    -------
    e : np.array
        Euclidean coordinates. Same shape as ``e`` except that the last
        last dimension is now has 3 elements.
    '''
    if np.all(h[..., 3] == 0) or np.allclose(h[..., 3], 1):
        return h[..., :3]
    elif np.all(h[..., 3] != 0):
        return (h[..., :3] / h[..., 3][..., None])
    else:
        raise ValueError('Input array must be either all euklidean points or all points at infinity.')


def distance_point_point(h_p1, h_p2):
    '''Calculate the Eukledian distance between 2 points

    Parameters
    ----------
    h_p1, h_p2 : np.ndarray of shape (4) or (N, 4)
        homogeneous coordiantes of input points

    Returns
    -------
    d : np.array of shape (1) or (N)
        Eukledian distance between those points
    '''
    if max(h_p1.ndim, h_p2.ndim) == 1:
        axis = 0
    elif max(h_p1.ndim, h_p2.ndim) == 2:
        axis = 1
    else:
        raise ValueError('This function expects 1d or 2d input.')
    return np.linalg.norm(h2e(h_p1) - h2e(h_p2), axis=axis)


def translation2aff(vec):
    '''Transform 3-d translation vector to affine 4*4 matrix.

    Parameters
    ----------
    vec : (3, ) array
        x, y, z translation vector

    Returns
    -------
    m : (4, 4) array
        affine transformation matrix
    '''
    if len(vec) != 3:
        raise ValueError('3d translation vector expected.')
    m = np.eye(4)
    m[:3, 3] = vec
    return m


def zoom2aff(zoom):
    '''Transform zoom to affine 4*4 matrix.

    Parameters
    ----------
    vec : float or (3, ) array
        zoom factor for x, y, z dimension (the same zoom is applied to
        every dimension is the input is a scalar).

    Returns
    -------
    m : (4, 4) array
        affine transformation matrix
    '''
    z = np.eye(4)
    if np.isscalar(zoom):
        z[:3,:3] = zoom * z[:3, :3]
    elif len(zoom) == 3:
        z[:3, :3] = np.diag(zoom)
    else:
        raise ValueError('Zoom must be either scalar or 3 element vector.')
    return z


def mat2aff(mat):
    '''Transform 3*3 matrix (e.g. rotation) to affine 4*4 matrix.

    Parameters
    ----------
    mat : (3, 3) array
        input matrix

    Returns
    -------
    m : (4, 4) array
        affine transformation matrix
    '''
    m = np.eye(4)
    m[:3, :3] = mat
    return m

def norm_vector(vec):
    '''Normalize euklidean vectors.

    Parameters
    ----------
    vec : np.array
        Input vectors of shape (n, 3)

    Returns
    -------
    vec : np.array
        Normalized vectors
    '''
    length2 = np.sum(vec * vec, axis=-1)
    return vec / np.sqrt(length2)[:, None]

def anglediff(phi):
    '''Angle range covered by phi, accounting for 2 pi properly

    Parameters
    ----------
    phi : list of two float
        Two angles in radian
    '''
    anglediff = phi[1] - phi[0]
    if (anglediff < 0.) or (anglediff > (2. * np.pi)):
        # If anglediff == 2 pi exactly, presumably the user want to cover the full circle.
        anglediff = anglediff % (2. * np.pi)
    return anglediff

def angle_between(angle, border1, border2):
    '''Test if an angle is between two others

    Since angles are cyclic, a simple numerical comparison is not
    sufficient, e.g. 355 deg is in the interval [350 deg, 10 deg] even though
    numerically 355 is not less then 10.

    Parameters
    ----------
    angle : float or np.array
        Angle array (in rad)
    border1 : float
        Lower border of interval (inclusive) in radian
    border2 : float
        Higher border of interval (inclusive) in radian

    Returns
    -------
    comparison : bool of same shape as ``angle``
        Result of the comparison

    Examples
    --------

    >>> from marxs.math.utils import angle_between
    >>> angle_between(-0.1, -0.2, 6)
    True
    '''
    twopi = 2 * np.pi
    b1 = (twopi + (border1 % twopi)) % twopi
    b2 = (twopi + (border2 % twopi)) % twopi
    ang = (twopi + (angle % twopi)) % twopi
    if b1 < b2:
        return (b1 <= ang) & (ang <= b2)
    else:
        return (b1 <= ang) | (ang <= b2)



# Licensed under GPL version 3 - see LICENSE.rst
from __future__ import division
import warnings

import numpy as np
from transforms3d.affines import decompose44

from .base import FlatOpticalElement, OpticalElement
from ..utils import SimulationSetupWarning
from ..math.geometry import Cylinder

class FlatDetector(FlatOpticalElement):
    '''Flat detector with square pixels

    Processing the a photon with this detector adds four columns to the photon
    properties:

    - ``det_x``, ``det_y``: Event position on the detector in mm
    - ``detpix_x``, ``detpix_y``: Event position in detector coordinates, where
      (0, 0) is on the corner of the chip.

    The (x,y) coordinate system on the chip is such that it falls on the (y,z)
    plane of the global x,y,z coordinate system (it would be more logical to
    call the chip system (y,z), but traditionally that is not how chip
    coordinates are named). The pixel in the corner has coordinates (0, 0) in
    the pixel center.

    Parameters
    ----------
    pixsize : float
        size of pixels in mm

    ignore_pixel_warning : bool
        ignore warning if derived pixel number is not close to an integer

    kwargs :
       see `args for optical elements`

    '''

    loc_coos_name = ['det_x', 'det_y']
    '''name for output columns that contain the interaction point in local coordinates.'''

    detpix_name = ['detpix_x', 'detpix_y']
    '''name for output columns that contain this pixel number.'''



    display = {'color': (1.0, 1.0, 0.),
               'shape': 'box',
               'box-half': '+x'}

    def __init__(self, pixsize=1, ignore_pixel_warning=False, **kwargs):
        self.pixsize = pixsize
        super(FlatDetector, self).__init__(**kwargs)
        t, r, zoom, s = decompose44(self.pos4d)
        self.npix = [0, 0]
        self.centerpix = [0, 0]
        for i in (0, 1):
            z  = zoom[i + 1]
            self.npix[i] = int(np.round(2. * z / self.pixsize))
            if (np.abs(2. * z / self.pixsize - self.npix[i]) > 1e-2) and not ignore_pixel_warning:
                warnings.warn('Detector size is not an integer multiple of pixel size in direction {0}. It will be rounded.'.format('xy'[i]), SimulationSetupWarning)
            self.centerpix[i] = (self.npix[i] - 1) / 2

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        detx = intercoos[intersect, 0] / self.pixsize + self.centerpix[0]
        dety = intercoos[intersect, 1] / self.pixsize + self.centerpix[1]
        return {self.detpix_name[0]: detx, self.detpix_name[1]: dety}

class CircularDetector(OpticalElement):
    '''A detector shaped like a ring or tube.

    This detector is shaped like a tube. The form is a circle in the xy plane
    and and flat along the z-direction.  While most CCDs are flat in practice,
    the `CircularDetector` simulates a setup that can follow the Rowland circle
    geometry exactly which is useful, e.g. to study the resolution of a
    spectrograph without worrying about the details of the detector geometry.

    Parameters
    ----------
    position, orientation, zoom, pos4d : see description of `pos4d`
        The radius of the tube is given by the ``zoom`` keyword, see `pos4d`.
        Use ``zoom[0] == zoom[1]`` to make a circular tube. ``zoom[0] != zoom[1]`` gives
        an elliptical profile. ``zoom[2]`` sets the extension in the z direction.
    pixsize : float
        size of pixels in mm
    phi_offset : float
        This defines the center (pixel = 0) in radian. In other words:
        The output ``det_phi`` column has its zero position at ``phi_offset`` and
        values for ``phi`` are in the range [-pi, pi].

    '''
    loc_coos_name = ['det_phi', 'det_y']

    detpix_name = ['detpix_x', 'detpix_y']
    '''name for output columns that contain this pixel number.'''

    display = {'color': (1.0, 1.0, 0.),
               'opacity': 0.7,
               }

    centerpix = [0, 0]

    default_geometry = Cylinder

    def __init__(self, pixsize=1, **kwargs):
        self.pixsize = pixsize
        super(CircularDetector, self).__init__(**kwargs)

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        detx = intercoos[intersect, 0] * self.geometry['R'] / self.pixsize + self.centerpix[0]
        dety = intercoos[intersect, 1] / self.pixsize + self.centerpix[1]
        return {self.detpix_name[0]: detx, self.detpix_name[1]: dety}
